# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""ReportCapableInterfaces for registration tools."""
import os
from distutils.version import LooseVersion

from nipype.utils.filemanip import fname_presuffix
from nipype.interfaces.base import (
    traits,
    isdefined,
    File,
)
from nipype.interfaces.mixins import reporting
from nipype.interfaces import freesurfer as fs
from nipype.interfaces import fsl, ants

from ... import NIWORKFLOWS_LOG
from . import base as nrb
from ..norm import (
    _SpatialNormalizationInputSpec,
    _SpatialNormalizationOutputSpec,
    SpatialNormalization,
)
from ..fixes import (
    FixHeaderApplyTransforms as ApplyTransforms,
    FixHeaderRegistration as Registration,
)


class _SpatialNormalizationInputSpecRPT(
    nrb._SVGReportCapableInputSpec, _SpatialNormalizationInputSpec
):
    pass


class _SpatialNormalizationOutputSpecRPT(
    reporting.ReportCapableOutputSpec, _SpatialNormalizationOutputSpec
):
    pass


class SpatialNormalizationRPT(nrb.RegistrationRC, SpatialNormalization):
    input_spec = _SpatialNormalizationInputSpecRPT
    output_spec = _SpatialNormalizationOutputSpecRPT

    def _post_run_hook(self, runtime):
        # We need to dig into the internal ants.Registration interface
        self._fixed_image = self._get_ants_args()["fixed_image"]
        if isinstance(self._fixed_image, (list, tuple)):
            self._fixed_image = self._fixed_image[0]  # get first item if list

        if self._get_ants_args().get("fixed_image_mask") is not None:
            self._fixed_image_mask = self._get_ants_args().get("fixed_image_mask")
        self._moving_image = self.aggregate_outputs(runtime=runtime).warped_image
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(SpatialNormalizationRPT, self)._post_run_hook(runtime)


class _ANTSRegistrationInputSpecRPT(
    nrb._SVGReportCapableInputSpec, ants.registration.RegistrationInputSpec
):
    pass


class _ANTSRegistrationOutputSpecRPT(
    reporting.ReportCapableOutputSpec, ants.registration.RegistrationOutputSpec
):
    pass


class ANTSRegistrationRPT(nrb.RegistrationRC, Registration):
    input_spec = _ANTSRegistrationInputSpecRPT
    output_spec = _ANTSRegistrationOutputSpecRPT

    def _post_run_hook(self, runtime):
        self._fixed_image = self.inputs.fixed_image[0]
        self._moving_image = self.aggregate_outputs(runtime=runtime).warped_image
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(ANTSRegistrationRPT, self)._post_run_hook(runtime)


class _ANTSApplyTransformsInputSpecRPT(
    nrb._SVGReportCapableInputSpec, ants.resampling.ApplyTransformsInputSpec
):
    pass


class _ANTSApplyTransformsOutputSpecRPT(
    reporting.ReportCapableOutputSpec, ants.resampling.ApplyTransformsOutputSpec
):
    pass


class ANTSApplyTransformsRPT(nrb.RegistrationRC, ApplyTransforms):
    input_spec = _ANTSApplyTransformsInputSpecRPT
    output_spec = _ANTSApplyTransformsOutputSpecRPT

    def _post_run_hook(self, runtime):
        self._fixed_image = self.inputs.reference_image
        self._moving_image = self.aggregate_outputs(runtime=runtime).output_image
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(ANTSApplyTransformsRPT, self)._post_run_hook(runtime)


class _ApplyTOPUPInputSpecRPT(
    nrb._SVGReportCapableInputSpec, fsl.epi.ApplyTOPUPInputSpec
):
    wm_seg = File(argstr="-wmseg %s", desc="reference white matter segmentation mask")


class _ApplyTOPUPOutputSpecRPT(
    reporting.ReportCapableOutputSpec, fsl.epi.ApplyTOPUPOutputSpec
):
    pass


class ApplyTOPUPRPT(nrb.RegistrationRC, fsl.ApplyTOPUP):
    input_spec = _ApplyTOPUPInputSpecRPT
    output_spec = _ApplyTOPUPOutputSpecRPT

    def _post_run_hook(self, runtime):
        from nilearn.image import index_img

        self._fixed_image_label = "after"
        self._moving_image_label = "before"
        self._fixed_image = index_img(
            self.aggregate_outputs(runtime=runtime).out_corrected, 0
        )
        self._moving_image = index_img(self.inputs.in_files[0], 0)
        self._contour = self.inputs.wm_seg if isdefined(self.inputs.wm_seg) else None
        NIWORKFLOWS_LOG.info(
            "Report - setting corrected (%s) and warped (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(ApplyTOPUPRPT, self)._post_run_hook(runtime)


class _FUGUEInputSpecRPT(nrb._SVGReportCapableInputSpec, fsl.preprocess.FUGUEInputSpec):
    wm_seg = File(argstr="-wmseg %s", desc="reference white matter segmentation mask")


class _FUGUEOutputSpecRPT(
    reporting.ReportCapableOutputSpec, fsl.preprocess.FUGUEOutputSpec
):
    pass


class FUGUERPT(nrb.RegistrationRC, fsl.FUGUE):
    input_spec = _FUGUEInputSpecRPT
    output_spec = _FUGUEOutputSpecRPT

    def _post_run_hook(self, runtime):
        self._fixed_image_label = "after"
        self._moving_image_label = "before"
        self._fixed_image = self.aggregate_outputs(runtime=runtime).unwarped_file
        self._moving_image = self.inputs.in_file
        self._contour = self.inputs.wm_seg if isdefined(self.inputs.wm_seg) else None
        NIWORKFLOWS_LOG.info(
            "Report - setting corrected (%s) and warped (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(FUGUERPT, self)._post_run_hook(runtime)


class _FLIRTInputSpecRPT(nrb._SVGReportCapableInputSpec, fsl.preprocess.FLIRTInputSpec):
    pass


class _FLIRTOutputSpecRPT(
    reporting.ReportCapableOutputSpec, fsl.preprocess.FLIRTOutputSpec
):
    pass


class FLIRTRPT(nrb.RegistrationRC, fsl.FLIRT):
    input_spec = _FLIRTInputSpecRPT
    output_spec = _FLIRTOutputSpecRPT

    def _post_run_hook(self, runtime):
        self._fixed_image = self.inputs.reference
        self._moving_image = self.aggregate_outputs(runtime=runtime).out_file
        self._contour = self.inputs.wm_seg if isdefined(self.inputs.wm_seg) else None
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(FLIRTRPT, self)._post_run_hook(runtime)


class _ApplyXFMInputSpecRPT(
    nrb._SVGReportCapableInputSpec, fsl.preprocess.ApplyXFMInputSpec
):
    pass


class ApplyXFMRPT(FLIRTRPT, fsl.ApplyXFM):
    input_spec = _ApplyXFMInputSpecRPT
    output_spec = _FLIRTOutputSpecRPT


if LooseVersion("0.0.0") < fs.Info.looseversion() < LooseVersion("6.0.0"):
    _BBRegisterInputSpec = fs.preprocess.BBRegisterInputSpec
else:
    _BBRegisterInputSpec = fs.preprocess.BBRegisterInputSpec6


class _BBRegisterInputSpecRPT(nrb._SVGReportCapableInputSpec, _BBRegisterInputSpec):
    # Adds default=True, usedefault=True
    out_lta_file = traits.Either(
        traits.Bool,
        File,
        default=True,
        usedefault=True,
        argstr="--lta %s",
        min_ver="5.2.0",
        desc="write the transformation matrix in LTA format",
    )


class _BBRegisterOutputSpecRPT(
    reporting.ReportCapableOutputSpec, fs.preprocess.BBRegisterOutputSpec
):
    pass


class BBRegisterRPT(nrb.RegistrationRC, fs.BBRegister):
    input_spec = _BBRegisterInputSpecRPT
    output_spec = _BBRegisterOutputSpecRPT

    def _post_run_hook(self, runtime):
        outputs = self.aggregate_outputs(runtime=runtime)
        mri_dir = os.path.join(self.inputs.subjects_dir, self.inputs.subject_id, "mri")
        target_file = os.path.join(mri_dir, "brainmask.mgz")

        # Apply transform for simplicity
        mri_vol2vol = fs.ApplyVolTransform(
            source_file=self.inputs.source_file,
            target_file=target_file,
            lta_file=outputs.out_lta_file,
            interp="nearest",
        )
        res = mri_vol2vol.run()

        self._fixed_image = target_file
        self._moving_image = res.outputs.transformed_file
        self._contour = os.path.join(mri_dir, "ribbon.mgz")
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(BBRegisterRPT, self)._post_run_hook(runtime)


class _MRICoregInputSpecRPT(
    nrb._SVGReportCapableInputSpec, fs.registration.MRICoregInputSpec
):
    pass


class _MRICoregOutputSpecRPT(
    reporting.ReportCapableOutputSpec, fs.registration.MRICoregOutputSpec
):
    pass


class MRICoregRPT(nrb.RegistrationRC, fs.MRICoreg):
    input_spec = _MRICoregInputSpecRPT
    output_spec = _MRICoregOutputSpecRPT

    def _post_run_hook(self, runtime):
        outputs = self.aggregate_outputs(runtime=runtime)
        mri_dir = None
        if isdefined(self.inputs.subject_id):
            mri_dir = os.path.join(
                self.inputs.subjects_dir, self.inputs.subject_id, "mri"
            )

        if isdefined(self.inputs.reference_file):
            target_file = self.inputs.reference_file
        else:
            target_file = os.path.join(mri_dir, "brainmask.mgz")

        # Apply transform for simplicity
        mri_vol2vol = fs.ApplyVolTransform(
            source_file=self.inputs.source_file,
            target_file=target_file,
            lta_file=outputs.out_lta_file,
            interp="nearest",
        )
        res = mri_vol2vol.run()

        self._fixed_image = target_file
        self._moving_image = res.outputs.transformed_file
        if mri_dir is not None:
            self._contour = os.path.join(mri_dir, "ribbon.mgz")
        NIWORKFLOWS_LOG.info(
            "Report - setting fixed (%s) and moving (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(MRICoregRPT, self)._post_run_hook(runtime)


class _SimpleBeforeAfterInputSpecRPT(nrb._SVGReportCapableInputSpec):
    before = File(exists=True, mandatory=True, desc="file before")
    after = File(exists=True, mandatory=True, desc="file after")
    wm_seg = File(desc="reference white matter segmentation mask")
    before_label = traits.Str("before", usedefault=True)
    after_label = traits.Str("after", usedefault=True)
    dismiss_affine = traits.Bool(
        False, usedefault=True, desc="rotate image(s) to cardinal axes"
    )


class SimpleBeforeAfterRPT(nrb.RegistrationRC, nrb.ReportingInterface):
    input_spec = _SimpleBeforeAfterInputSpecRPT

    def _post_run_hook(self, runtime):
        """ there is not inner interface to run """
        self._fixed_image_label = self.inputs.after_label
        self._moving_image_label = self.inputs.before_label
        self._fixed_image = self.inputs.after
        self._moving_image = self.inputs.before
        self._contour = self.inputs.wm_seg if isdefined(self.inputs.wm_seg) else None
        self._dismiss_affine = self.inputs.dismiss_affine
        NIWORKFLOWS_LOG.info(
            "Report - setting before (%s) and after (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        return super(SimpleBeforeAfterRPT, self)._post_run_hook(runtime)


class _ResampleBeforeAfterInputSpecRPT(_SimpleBeforeAfterInputSpecRPT):
    base = traits.Enum("before", "after", usedefault=True, mandatory=True)


class ResampleBeforeAfterRPT(SimpleBeforeAfterRPT):
    input_spec = _ResampleBeforeAfterInputSpecRPT

    def _post_run_hook(self, runtime):
        from nilearn import image as nli

        self._fixed_image = self.inputs.after
        self._moving_image = self.inputs.before
        if self.inputs.base == "before":
            resampled_after = nli.resample_to_img(self._fixed_image, self._moving_image)
            fname = fname_presuffix(
                self._fixed_image, suffix="_resampled", newpath=runtime.cwd
            )
            resampled_after.to_filename(fname)
            self._fixed_image = fname
        else:
            resampled_before = nli.resample_to_img(
                self._moving_image, self._fixed_image
            )
            fname = fname_presuffix(
                self._moving_image, suffix="_resampled", newpath=runtime.cwd
            )
            resampled_before.to_filename(fname)
            self._moving_image = fname
        self._contour = self.inputs.wm_seg if isdefined(self.inputs.wm_seg) else None
        NIWORKFLOWS_LOG.info(
            "Report - setting before (%s) and after (%s) images",
            self._fixed_image,
            self._moving_image,
        )

        runtime = super(ResampleBeforeAfterRPT, self)._post_run_hook(runtime)
        NIWORKFLOWS_LOG.info("Successfully created report (%s)", self._out_report)
        os.unlink(fname)

        return runtime



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Select terms for a confound model, and compute any requisite expansions."""
import os
import re
import numpy as np
import pandas as pd
from functools import reduce
from collections import deque, OrderedDict
from nipype.utils.filemanip import fname_presuffix
from nipype.utils.misc import normalize_mc_params
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterfaceInputSpec,
    File,
    isdefined,
    SimpleInterface,
)


class _NormalizeMotionParamsInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input parameters file")
    format = traits.Enum(
        "FSL", "AFNI", "FSFAST", "NIPY", usedefault=True, desc="output format"
    )


class _NormalizeMotionParamsOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="written file path")


class NormalizeMotionParams(SimpleInterface):
    """Convert input motion parameters into the designated convention."""

    input_spec = _NormalizeMotionParamsInputSpec
    output_spec = _NormalizeMotionParamsOutputSpec

    def _run_interface(self, runtime):
        mpars = np.loadtxt(self.inputs.in_file)  # mpars is N_t x 6
        mpars = np.apply_along_axis(
            func1d=normalize_mc_params, axis=1, arr=mpars, source=self.inputs.format
        )
        self._results["out_file"] = os.path.join(runtime.cwd, "motion_params.txt")
        np.savetxt(self._results["out_file"], mpars)
        return runtime


class _ExpandModelInputSpec(BaseInterfaceInputSpec):
    confounds_file = File(
        exists=True,
        mandatory=True,
        desc="TSV containing confound time series for "
        "expansion according to the specified formula.",
    )
    model_formula = traits.Str(
        "(dd1(rps + wm + csf + gsr))^^2 + others",
        usedefault=True,
        desc="""\
Formula for generating model expansions. By default, the
32-parameter expansion will be generated. Note that any expressions
to be expanded *must* be in parentheses, even if they include only
a single variable (e.g., ``(x)^2``, not ``x^2``).

Examples:

  * ``rps + wm + csf + gsr`` : 9-parameter model. ``rps`` denotes realignment
    parameters, ``wm`` denotes mean white matter signal, ``csf`` denotes mean
    cerebrospinal fluid signal, and ``gsr`` denotes mean global signal.
  * ``(dd1(rps + wm + csf + gsr))^^2`` : 36-parameter expansion.
    ``rps + wm + csf + gsr`` denotes that realignment parameters and mean
    WM, CSF, and global signals should be included. ``dd1`` denotes that
    these signals should be augmented with their first temporal
    derivatives. ``^^2`` denotes that the original signals and temporal
    derivatives should be augmented with quadratic expansions.
  * ``(dd1(rps))^^2`` : 24-parameter expansion. ``rps`` denotes that
    realignment parameters should be included. ``dd1`` and ``^^2`` denote
    temporal derivative and quadratic expansions as above.
  * ``(dd1(rps + wm + csf + gsr))^^2 + others`` : generate all expansion
    terms necessary for a 36-parameter model as above, and
    concatenate those expansion terms to all other regressor columns
    in the confounds file.

""",
    )
    output_file = File(desc="Output path")


class _ExpandModelOutputSpec(TraitedSpec):
    confounds_file = File(exists=True, desc="Output confounds file")


class ExpandModel(SimpleInterface):
    """Expand a confound model according to a specified formula."""

    input_spec = _ExpandModelInputSpec
    output_spec = _ExpandModelOutputSpec

    def _run_interface(self, runtime):
        if isdefined(self.inputs.output_file):
            out_file = self.inputs.output_file
        else:
            out_file = fname_presuffix(
                self.inputs.confounds_file,
                suffix="_expansion.tsv",
                newpath=runtime.cwd,
                use_ext=False,
            )

        confounds_data = pd.read_csv(self.inputs.confounds_file, sep="\t")
        _, confounds_data = parse_formula(
            model_formula=self.inputs.model_formula,
            parent_data=confounds_data,
            unscramble=True,
        )
        confounds_data.to_csv(out_file, sep="\t", index=False, na_rep="n/a")
        self._results["confounds_file"] = out_file
        return runtime


class _SpikeRegressorsInputSpec(BaseInterfaceInputSpec):
    confounds_file = File(
        exists=True,
        mandatory=True,
        desc="TSV containing criterion time series (e.g., framewise "
        "displacement, DVARS) to be used for creating spike regressors.",
    )
    fd_thresh = traits.Float(
        0.5,
        usedefault=True,
        desc="Minimum framewise displacement threshold for flagging a frame "
        "as a spike.",
    )
    dvars_thresh = traits.Float(
        1.5,
        usedefault=True,
        desc="Minimum standardised DVARS threshold for flagging a frame as " "a spike.",
    )
    header_prefix = traits.Str(
        "motion_outlier",
        usedefault=True,
        desc="Prefix for spikes in the output TSV header",
    )
    lags = traits.List(
        traits.Int,
        value=[0],
        usedefault=True,
        desc="Relative indices of lagging frames to flag for " "each flagged frame",
    )
    minimum_contiguous = traits.Either(
        None,
        traits.Int,
        usedefault=True,
        desc="Minimum number of contiguous volumes required to avoid "
        "flagging as a spike",
    )
    concatenate = traits.Bool(
        True,
        usedefault=True,
        desc="Indicates whether to concatenate spikes to existing confounds "
        "or return spikes only",
    )
    output_format = traits.Enum(
        "spikes", "mask", usedefault=True, desc="Format of output (spikes or mask)"
    )
    output_file = File(desc="Output path")


class _SpikeRegressorsOutputSpec(TraitedSpec):
    confounds_file = File(exists=True, desc="Output confounds file")


class SpikeRegressors(SimpleInterface):
    """Generate spike regressors."""

    input_spec = _SpikeRegressorsInputSpec
    output_spec = _SpikeRegressorsOutputSpec

    def _run_interface(self, runtime):
        if isdefined(self.inputs.output_file):
            out_file = self.inputs.output_file
        else:
            out_file = fname_presuffix(
                self.inputs.confounds_file,
                suffix="_desc-motion_outliers.tsv",
                newpath=runtime.cwd,
                use_ext=False,
            )

        spike_criteria = {
            "framewise_displacement": (">", self.inputs.fd_thresh),
            "std_dvars": (">", self.inputs.dvars_thresh),
        }

        confounds_data = pd.read_csv(self.inputs.confounds_file, sep="\t")
        confounds_data = spike_regressors(
            data=confounds_data,
            criteria=spike_criteria,
            header_prefix=self.inputs.header_prefix,
            lags=self.inputs.lags,
            minimum_contiguous=self.inputs.minimum_contiguous,
            concatenate=self.inputs.concatenate,
            output=self.inputs.output_format,
        )
        confounds_data.to_csv(out_file, sep="\t", index=False, na_rep="n/a")
        self._results["confounds_file"] = out_file
        return runtime


def spike_regressors(
    data,
    criteria=None,
    header_prefix="motion_outlier",
    lags=None,
    minimum_contiguous=None,
    concatenate=True,
    output="spikes",
):
    """
    Add spike regressors to a confound/nuisance matrix.

    Parameters
    ----------
    data : :obj:`~pandas.DataFrame`
        A tabulation of observations from which spike regressors should be
        estimated.
    criteria : :obj:`dict` of (:obj:`str`, ``'>'`` or ``'<'`` or :obj:`float`)
        Criteria for generating a spike regressor. If, for a given frame, the
        value of the variable corresponding to the key exceeds the threshold
        indicated by the value, then a spike regressor is created for that
        frame. By default, the strategy from [Power2014]_ is implemented: any
        frames with FD greater than 0.5 or standardised DV greater than 1.5
        are flagged for censoring.
    header_prefix : :obj:`str`
        The prefix used to indicate spike regressors in the output data table.
    lags: :obj:`list` of :obj:`int`
        A list indicating the frames to be censored relative to each flag.
        For instance, ``[0]`` censors the flagged frame, while ``[0, 1]`` censors
        both the flagged frame and the following frame.
    minimum_contiguous : :obj:`int` or :obj:`None`
        The minimum number of contiguous frames that must be unflagged for
        spike regression. If any series of contiguous unflagged frames is
        shorter than the specified minimum, then all of those frames will
        additionally have spike regressors implemented.
    concatenate : :obj:`bool`
        Indicates whether the returned object should include only spikes
        (if false) or all input time series and spikes (if true, default).
    output : :obj:`str`
        Indicates whether the output should be formatted as spike regressors
        ('spikes', a separate column for each outlier) or as a temporal mask
        ('mask', a single output column indicating the locations of outliers).

    Returns
    -------
    data : :obj:`~pandas.DataFrame`
        The input DataFrame with a column for each spike regressor.

    References
    ----------
    .. [Power2014] Power JD, et al. (2014)
        Methods to detect, characterize, and remove motion artifact in resting
        state fMRI. NeuroImage. doi:`10.1016/j.neuroimage.2013.08.048
        <https://doi.org/10.1016/j.neuroimage.2013.08.048>`__.

    """
    mask = {}
    indices = range(data.shape[0])
    lags = lags or [0]
    criteria = criteria or {
        "framewise_displacement": (">", 0.5),
        "std_dvars": (">", 1.5),
    }
    for metric, (criterion, threshold) in criteria.items():
        if criterion == "<":
            mask[metric] = set(np.where(data[metric] < threshold)[0])
        elif criterion == ">":
            mask[metric] = set(np.where(data[metric] > threshold)[0])
    mask = reduce((lambda x, y: x | y), mask.values())

    for lag in lags:
        mask = set([m + lag for m in mask]) | mask

    mask = mask.intersection(indices)
    if minimum_contiguous is not None:
        post_final = data.shape[0] + 1
        epoch_length = np.diff(sorted(mask | set([-1, post_final]))) - 1
        epoch_end = sorted(mask | set([post_final]))
        for end, length in zip(epoch_end, epoch_length):
            if length < minimum_contiguous:
                mask = mask | set(range(end - length, end))
        mask = mask.intersection(indices)

    if output == "mask":
        spikes = np.zeros(data.shape[0])
        spikes[list(mask)] = 1
        spikes = pd.DataFrame(data=spikes, columns=[header_prefix])
    else:
        spikes = np.zeros((max(indices) + 1, len(mask)))
        for i, m in enumerate(sorted(mask)):
            spikes[m, i] = 1
        header = ["{:s}{:02d}".format(header_prefix, vol) for vol in range(len(mask))]
        spikes = pd.DataFrame(data=spikes, columns=header)
    if concatenate:
        return pd.concat((data, spikes), axis=1)
    else:
        return spikes


def temporal_derivatives(order, variables, data):
    """
    Compute temporal derivative terms by the method of backwards differences.

    Parameters
    ----------
    order : :obj:`list` of :obj:`int`
        A list of temporal derivative terms to include. For instance, ``[1, 2]``
        indicates that the first and second derivative terms should be added.
        To retain the original terms, ``0`` *must* be included in the list.
    variables : :obj:`list` of :obj:`str`
        List of variables for which temporal derivative terms should be
        computed.
    data : :obj:`~pandas.DataFrame`
        Table of values of all observations of all variables.

    Returns
    -------
    variables_deriv : :obj:`list`
        A list of variables to include in the final data frame after adding
        the specified derivative terms.
    data_deriv : :obj:`~pandas.DataFrame`
        Table of values of all observations of all variables, including any
        specified derivative terms.

    """
    variables_deriv = OrderedDict()
    data_deriv = OrderedDict()
    if 0 in order:
        data_deriv[0] = data[variables]
        variables_deriv[0] = variables
        order = set(order) - set([0])
    for o in order:
        variables_deriv[o] = ["{}_derivative{}".format(v, o) for v in variables]
        data_deriv[o] = np.tile(np.nan, data[variables].shape)
        data_deriv[o][o:, :] = np.diff(data[variables], n=o, axis=0)
    variables_deriv = reduce((lambda x, y: x + y), variables_deriv.values())
    data_deriv = pd.DataFrame(
        columns=variables_deriv, data=np.concatenate([*data_deriv.values()], axis=1)
    )

    return (variables_deriv, data_deriv)


def exponential_terms(order, variables, data):
    """
    Compute exponential expansions.

    Parameters
    ----------
    order : :obj:`list` of :obj:`int`
        A list of exponential terms to include. For instance, ``[1, 2]``
        indicates that the first and second exponential terms should be added.
        To retain the original terms, ``1`` *must* be included in the list.
    variables : :obj:`list` of :obj:`str`
        List of variables for which exponential terms should be computed.
    data : :obj:`~pandas.DataFrame`
        Table of values of all observations of all variables.

    Returns
    -------
    variables_exp : :obj:`list`
        A list of variables to include in the final data frame after adding
        the specified exponential terms.
    data_exp : :obj:`~pandas.DataFrame`
        Table of values of all observations of all variables, including any
        specified exponential terms.

    """
    variables_exp = OrderedDict()
    data_exp = OrderedDict()
    if 1 in order:
        data_exp[1] = data[variables]
        variables_exp[1] = variables
        order = set(order) - set([1])
    for o in order:
        variables_exp[o] = ["{}_power{}".format(v, o) for v in variables]
        data_exp[o] = data[variables] ** o
    variables_exp = reduce((lambda x, y: x + y), variables_exp.values())
    data_exp = pd.DataFrame(
        columns=variables_exp, data=np.concatenate([*data_exp.values()], axis=1)
    )
    return (variables_exp, data_exp)


def _order_as_range(order):
    """Convert a hyphenated string representing order for derivative or
    exponential terms into a range object that can be passed as input to the
    appropriate expansion function."""
    order = order.split("-")
    order = [int(o) for o in order]
    if len(order) > 1:
        order = range(order[0], (order[-1] + 1))
    return order


def _check_and_expand_exponential(expr, variables, data):
    """Check if the current operation specifies exponential expansion. ^^6
    specifies all powers up to the 6th, ^5-6 the 5th and 6th powers, ^6 the
    6th only."""
    if re.search(r"\^\^[0-9]+$", expr):
        order = re.compile(r"\^\^([0-9]+)$").findall(expr)
        order = range(1, int(*order) + 1)
        variables, data = exponential_terms(order, variables, data)
    elif re.search(r"\^[0-9]+[\-]?[0-9]*$", expr):
        order = re.compile(r"\^([0-9]+[\-]?[0-9]*)").findall(expr)
        order = _order_as_range(*order)
        variables, data = exponential_terms(order, variables, data)
    return variables, data


def _check_and_expand_derivative(expr, variables, data):
    """Check if the current operation specifies a temporal derivative. dd6x
    specifies all derivatives up to the 6th, d5-6x the 5th and 6th, d6x the
    6th only."""
    if re.search(r"^dd[0-9]+", expr):
        order = re.compile(r"^dd([0-9]+)").findall(expr)
        order = range(0, int(*order) + 1)
        (variables, data) = temporal_derivatives(order, variables, data)
    elif re.search(r"^d[0-9]+[\-]?[0-9]*", expr):
        order = re.compile(r"^d([0-9]+[\-]?[0-9]*)").findall(expr)
        order = _order_as_range(*order)
        (variables, data) = temporal_derivatives(order, variables, data)
    return variables, data


def _check_and_expand_subformula(expression, parent_data, variables, data):
    """Check if the current operation contains a suboperation, and parse it
    where appropriate."""
    grouping_depth = 0
    for i, char in enumerate(expression):
        if char == "(":
            if grouping_depth == 0:
                formula_delimiter = i + 1
            grouping_depth += 1
        elif char == ")":
            grouping_depth -= 1
            if grouping_depth == 0:
                expr = expression[formula_delimiter:i].strip()
                return parse_formula(expr, parent_data)
    return variables, data


def parse_expression(expression, parent_data):
    """
    Parse an expression in a model formula.

    Parameters
    ----------
    expression : :obj:`str`
        Formula expression: either a single variable or a variable group
        paired with an operation (exponentiation or differentiation).
    parent_data : :obj:`~pandas.DataFrame`
        The source data for the model expansion.

    Returns
    -------
    variables : :obj:`list`
        A list of variables in the provided formula expression.
    data: :obj:`~pandas.DataFrame`
        A tabulation of all terms in the provided formula expression.

    """
    variables = None
    data = None
    variables, data = _check_and_expand_subformula(
        expression, parent_data, variables, data
    )
    variables, data = _check_and_expand_exponential(expression, variables, data)
    variables, data = _check_and_expand_derivative(expression, variables, data)
    if variables is None:
        expr = expression.strip()
        variables = [expr]
        data = parent_data[expr]
    return variables, data


def _get_matches_from_data(regex, variables):
    matches = re.compile(regex)
    matches = " + ".join([v for v in variables if matches.match(v)])
    return matches


def _get_variables_from_formula(model_formula):
    symbols_to_clear = [
        " ",
        r"\(",
        r"\)",
        "dd[0-9]+",
        r"d[0-9]+[\-]?[0-9]*",
        r"\^\^[0-9]+",
        r"\^[0-9]+[\-]?[0-9]*",
    ]
    for symbol in symbols_to_clear:
        model_formula = re.sub(symbol, "", model_formula)
    variables = model_formula.split("+")
    return variables


def _expand_shorthand(model_formula, variables):
    """Expand shorthand terms in the model formula."""
    wm = "white_matter"
    gsr = "global_signal"
    rps = "trans_x + trans_y + trans_z + rot_x + rot_y + rot_z"
    fd = "framewise_displacement"
    acc = _get_matches_from_data("a_comp_cor_[0-9]+", variables)
    tcc = _get_matches_from_data("t_comp_cor_[0-9]+", variables)
    dv = _get_matches_from_data("^std_dvars$", variables)
    dvall = _get_matches_from_data(".*dvars", variables)
    nss = _get_matches_from_data("non_steady_state_outlier[0-9]+", variables)
    spikes = _get_matches_from_data("motion_outlier[0-9]+", variables)

    model_formula = re.sub("wm", wm, model_formula)
    model_formula = re.sub("gsr", gsr, model_formula)
    model_formula = re.sub("rps", rps, model_formula)
    model_formula = re.sub("fd", fd, model_formula)
    model_formula = re.sub("acc", acc, model_formula)
    model_formula = re.sub("tcc", tcc, model_formula)
    model_formula = re.sub("dv", dv, model_formula)
    model_formula = re.sub("dvall", dvall, model_formula)
    model_formula = re.sub("nss", nss, model_formula)
    model_formula = re.sub("spikes", spikes, model_formula)

    formula_variables = _get_variables_from_formula(model_formula)
    others = " + ".join(set(variables) - set(formula_variables))
    model_formula = re.sub("others", others, model_formula)
    return model_formula


def _unscramble_regressor_columns(parent_data, data):
    """Reorder the columns of a confound matrix such that the columns are in
    the same order as the input data with any expansion columns inserted
    immediately after the originals.
    """
    matches = ["_power[0-9]+", "_derivative[0-9]+"]
    var = OrderedDict((c, deque()) for c in parent_data.columns)
    for c in data.columns:
        col = c
        for m in matches:
            col = re.sub(m, "", col)
        if col == c:
            var[col].appendleft(c)
        else:
            var[col].append(c)
    unscrambled = reduce((lambda x, y: x + y), var.values())
    return data[[*unscrambled]]


def parse_formula(model_formula, parent_data, unscramble=False):
    """
    Parse a confound manipulation formula.

    Recursively parse a model formula by breaking it into additive atoms
    and tracking grouping symbol depth.

    Parameters
    ----------
    model_formula : :obj:`str`
        Expression for the model formula, e.g.
        ``(a + b)^^2 + dd1(c + (d + e)^3) + f``
        Note that any expressions to be expanded *must* be in parentheses,
        even if they include only a single variable
        (e.g., ``(x)^2``, not ``x^2``).
        Temporal derivative options:

            * ``d6(variable)`` for the 6th temporal derivative
            * ``dd6(variable)`` for all temporal derivatives up to the 6th
            * ``d4-6(variable)`` for the 4th through 6th temporal derivatives
            * ``0`` must be included in the temporal derivative range for the
              original term to be returned when temporal derivatives are computed.

        Exponential options:

            * ``(variable)^6`` for the 6th power
            * ``(variable)^^6`` for all powers up to the 6th
            * ``(variable)^4-6`` for the 4th through 6th powers
            * ``1`` must be included in the powers range for the original term to be
              returned when exponential terms are computed.

        Temporal derivatives and exponential terms are computed for all terms
        in the grouping symbols that they adjoin.
    parent_data : :obj:`~pandas.DataFrame`
        A tabulation of all values usable in the model formula. Each additive
        term in `model_formula` should correspond either to a variable in this
        data frame or to instructions for operating on a variable (for
        instance, computing temporal derivatives or exponential terms).

    Returns
    -------
    variables : :obj:`list` of :obj:`str`
        A list of variables included in the model parsed from the provided
        formula.
    data : :obj:`~pandas.DataFrame`
        All values in the complete model.

    """
    variables = {}
    data = {}
    expr_delimiter = 0
    grouping_depth = 0
    model_formula = _expand_shorthand(model_formula, parent_data.columns)
    for i, char in enumerate(model_formula):
        if char == "(":
            grouping_depth += 1
        elif char == ")":
            grouping_depth -= 1
        elif grouping_depth == 0 and char == "+":
            expression = model_formula[expr_delimiter:i].strip()
            variables[expression] = None
            data[expression] = None
            expr_delimiter = i + 1
    expression = model_formula[expr_delimiter:].strip()
    variables[expression] = None
    data[expression] = None
    for expression in list(variables):
        if expression[0] == "(" and expression[-1] == ")":
            (variables[expression], data[expression]) = parse_formula(
                expression[1:-1], parent_data
            )
        else:
            (variables[expression], data[expression]) = parse_expression(
                expression, parent_data
            )
    variables = list(set(reduce((lambda x, y: x + y), variables.values())))
    data = pd.concat((data.values()), axis=1)

    if unscramble:
        data = _unscramble_regressor_columns(parent_data, data)

    return variables, data



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Reports builder for BIDS-Apps.

Generalizes report generation across BIDS-Apps

"""
from pathlib import Path
import re
from itertools import compress
from collections import defaultdict
from pkg_resources import resource_filename as pkgrf
from bids.layout import BIDSLayout, add_config_paths
import jinja2
from nipype.utils.filemanip import copyfile

# Add a new figures spec
try:
    add_config_paths(figures=pkgrf("niworkflows", "data/nipreps.json"))
except ValueError as e:
    if "Configuration 'figures' already exists" != str(e):
        raise

PLURAL_SUFFIX = defaultdict(str("s").format, [("echo", "es")])
SVG_SNIPPET = [
    """\
<object class="svg-reportlet" type="image/svg+xml" data="./{0}">
Problem loading figure {0}. If the link below works, please try \
reloading the report in your browser.</object>
</div>
<div class="elem-filename">
    Get figure file: <a href="./{0}" target="_blank">{0}</a>
</div>
""",
    """\
<img class="svg-reportlet" src="./{0}" style="width: 100%" />
</div>
<div class="elem-filename">
    Get figure file: <a href="./{0}" target="_blank">{0}</a>
</div>
""",
]


class Element(object):
    """Just a basic component of a report"""

    def __init__(self, name, title=None):
        self.name = name
        self.title = title


class Reportlet(Element):
    """
    A reportlet has title, description and a list of components with either an
    HTML fragment or a path to an SVG file, and possibly a caption. This is a
    factory class to generate Reportlets reusing the layout from a ``Report``
    object.

    .. testsetup::

    >>> cwd = os.getcwd()
    >>> os.chdir(tmpdir)

    >>> from pkg_resources import resource_filename
    >>> from shutil import copytree
    >>> from bids.layout import BIDSLayout
    >>> test_data_path = resource_filename('niworkflows', 'data/tests/work')
    >>> testdir = Path(tmpdir)
    >>> data_dir = copytree(test_data_path, str(testdir / 'work'))
    >>> out_figs = testdir / 'out' / 'fmriprep'
    >>> bl = BIDSLayout(str(testdir / 'work' / 'reportlets'),
    ...                 config='figures', validate=False)

    .. doctest::

    >>> bl.get(subject='01', desc='reconall') # doctest: +ELLIPSIS
    [<BIDSFile filename='.../fmriprep/sub-01/figures/sub-01_desc-reconall_T1w.svg'>]

    >>> len(bl.get(subject='01', space='.*', regex_search=True))
    2

    >>> r = Reportlet(bl, out_figs, config={
    ...     'title': 'Some Title', 'bids': {'datatype': 'figures', 'desc': 'reconall'},
    ...     'description': 'Some description'})
    >>> r.name
    'datatype-figures_desc-reconall'

    >>> r.components[0][0].startswith('<img')
    True

    >>> r = Reportlet(bl, out_figs, config={
    ...     'title': 'Some Title', 'bids': {'datatype': 'figures', 'desc': 'reconall'},
    ...     'description': 'Some description', 'static': False})
    >>> r.name
    'datatype-figures_desc-reconall'

    >>> r.components[0][0].startswith('<object')
    True

    >>> r = Reportlet(bl, out_figs, config={
    ...     'title': 'Some Title', 'bids': {'datatype': 'figures', 'desc': 'summary'},
    ...     'description': 'Some description'})

    >>> r.components[0][0].startswith('<h3')
    True

    >>> r.components[0][1] is None
    True

    >>> r = Reportlet(bl, out_figs, config={
    ...     'title': 'Some Title',
    ...     'bids': {'datatype': 'figures', 'space': '.*', 'regex_search': True},
    ...     'caption': 'Some description {space}'})
    >>> sorted(r.components)[0][1]
    'Some description MNI152NLin2009cAsym'

    >>> sorted(r.components)[1][1]
    'Some description MNI152NLin6Asym'


    >>> r = Reportlet(bl, out_figs, config={
    ...     'title': 'Some Title',
    ...     'bids': {'datatype': 'fmap', 'space': '.*', 'regex_search': True},
    ...     'caption': 'Some description {space}'})
    >>> r.is_empty()
    True

    .. testcleanup::

    >>> os.chdir(cwd)

    """

    def __init__(self, layout, out_dir, config=None):
        if not config:
            raise RuntimeError("Reportlet must have a config object")

        self.name = config.get(
            "name", "_".join("%s-%s" % i for i in sorted(config["bids"].items()))
        )
        self.title = config.get("title")
        self.subtitle = config.get("subtitle")
        self.description = config.get("description")

        # Query the BIDS layout of reportlets
        files = layout.get(**config["bids"])

        self.components = []
        for bidsfile in files:
            src = Path(bidsfile.path)
            ext = "".join(src.suffixes)
            desc_text = config.get("caption")

            contents = None
            if ext == ".html":
                contents = src.read_text().strip()
            elif ext == ".svg":
                entities = dict(bidsfile.entities)
                if desc_text:
                    desc_text = desc_text.format(**entities)

                try:
                    html_anchor = src.relative_to(out_dir)
                except ValueError:
                    html_anchor = src.relative_to(Path(layout.root).parent)
                    dst = out_dir / html_anchor
                    dst.parent.mkdir(parents=True, exist_ok=True)
                    copyfile(src, dst, copy=True, use_hardlink=True)

                contents = SVG_SNIPPET[config.get("static", True)].format(html_anchor)

                # Our current implementations of dynamic reportlets do this themselves,
                # however I'll leave the code here since this is potentially something we
                # will want to transfer from every figure generator to this location.
                # The following code misses setting preserveAspecRatio="xMidYMid meet"
                # if not is_static:
                #     # Remove height and width attributes from initial <svg> tag
                #     svglines = out_file.read_text().splitlines()
                #     expr = re.compile(r' (height|width)=["\'][0-9]+(\.[0-9]*)?[a-z]*["\']')
                #     for l, line in enumerate(svglines[:6]):
                #         if line.strip().startswith('<svg'):
                #             newline = expr.sub('', line)
                #             svglines[l] = newline
                #             out_file.write_text('\n'.join(svglines))
                #             break

            if contents:
                self.components.append((contents, desc_text))

    def is_empty(self):
        return len(self.components) == 0


class SubReport(Element):
    """SubReports are sections within a Report."""

    def __init__(self, name, isnested=False, reportlets=None, title=""):
        self.name = name
        self.title = title
        self.reportlets = reportlets or []
        self.isnested = isnested


class Report:
    """
    The full report object. This object maintains a BIDSLayout to index
    all reportlets.


    .. testsetup::

    >>> cwd = os.getcwd()
    >>> os.chdir(tmpdir)

    >>> from pkg_resources import resource_filename
    >>> from shutil import copytree
    >>> from bids.layout import BIDSLayout
    >>> test_data_path = resource_filename('niworkflows', 'data/tests/work')
    >>> testdir = Path(tmpdir)
    >>> data_dir = copytree(test_data_path, str(testdir / 'work'))
    >>> out_figs = testdir / 'out' / 'fmriprep'

    .. doctest::

    >>> robj = Report(testdir / 'out', 'madeoutuuid', subject_id='01', packagename='fmriprep',
    ...               reportlets_dir=testdir / 'work' / 'reportlets')
    >>> robj.layout.get(subject='01', desc='reconall')  # doctest: +ELLIPSIS
    [<BIDSFile filename='.../figures/sub-01_desc-reconall_T1w.svg'>]

    >>> robj.generate_report()
    0
    >>> len((testdir / 'out' / 'fmriprep' / 'sub-01.html').read_text())
    36693

    .. testcleanup::

    >>> os.chdir(cwd)

    """

    def __init__(
        self,
        out_dir,
        run_uuid,
        config=None,
        out_filename="report.html",
        packagename=None,
        reportlets_dir=None,
        subject_id=None,
    ):
        self.root = Path(reportlets_dir or out_dir)

        # Initialize structuring elements
        self.sections = []
        self.errors = []
        self.out_dir = Path(out_dir)
        self.out_filename = out_filename
        self.run_uuid = run_uuid
        self.packagename = packagename
        self.subject_id = subject_id
        if subject_id is not None:
            self.subject_id = (
                subject_id[4:] if subject_id.startswith("sub-") else subject_id
            )
            self.out_filename = f"sub-{self.subject_id}.html"

        # Default template from niworkflows
        self.template_path = Path(pkgrf("niworkflows", "reports/report.tpl"))
        self._load_config(Path(config or pkgrf("niworkflows", "reports/default.yml")))
        assert self.template_path.exists()

    def _load_config(self, config):
        from yaml import safe_load as load

        settings = load(config.read_text())
        self.packagename = self.packagename or settings.get("package", None)

        if self.packagename is not None:
            self.root = self.root / self.packagename
            self.out_dir = self.out_dir / self.packagename

        if self.subject_id is not None:
            self.root = self.root / "sub-{}".format(self.subject_id)

        if "template_path" in settings:
            self.template_path = config.parent / settings["template_path"]

        self.index(settings["sections"])

    def init_layout(self):
        self.layout = BIDSLayout(self.root, config="figures", validate=False)

    def index(self, config):
        """
        Traverse the reports config definition and instantiate reportlets.

        This method also places figures in their final location.
        """
        # Initialize a BIDS layout
        self.init_layout()
        for subrep_cfg in config:
            # First determine whether we need to split by some ordering
            # (ie. sessions / tasks / runs), which are separated by commas.
            orderings = [
                s for s in subrep_cfg.get("ordering", "").strip().split(",") if s
            ]
            entities, list_combos = self._process_orderings(orderings, self.layout)

            if not list_combos:  # E.g. this is an anatomical reportlet
                reportlets = [
                    Reportlet(self.layout, self.out_dir, config=cfg)
                    for cfg in subrep_cfg["reportlets"]
                ]
            else:
                # Do not use dictionary for queries, as we need to preserve ordering
                # of ordering columns.
                reportlets = []
                for c in list_combos:
                    # do not display entities with the value None.
                    c_filt = list(filter(None, c))
                    ent_filt = list(compress(entities, c))
                    # Set a common title for this particular combination c
                    title = "Reports for: %s." % ", ".join(
                        [
                            '%s <span class="bids-entity">%s</span>'
                            % (ent_filt[i], c_filt[i])
                            for i in range(len(c_filt))
                        ]
                    )
                    for cfg in subrep_cfg["reportlets"]:
                        cfg["bids"].update({entities[i]: c[i] for i in range(len(c))})
                        rlet = Reportlet(self.layout, self.out_dir, config=cfg)
                        if not rlet.is_empty():
                            rlet.title = title
                            title = None
                            reportlets.append(rlet)

            # Filter out empty reportlets
            reportlets = [r for r in reportlets if not r.is_empty()]
            if reportlets:
                sub_report = SubReport(
                    subrep_cfg["name"],
                    isnested=bool(list_combos),
                    reportlets=reportlets,
                    title=subrep_cfg.get("title"),
                )
                self.sections.append(sub_report)

        # Populate errors section
        error_dir = (
            self.out_dir / "sub-{}".format(self.subject_id) / "log" / self.run_uuid
        )
        if error_dir.is_dir():
            from ..utils.misc import read_crashfile

            self.errors = [read_crashfile(str(f)) for f in error_dir.glob("crash*.*")]

    def generate_report(self):
        """Once the Report has been indexed, the final HTML can be generated"""
        logs_path = self.out_dir / "logs"

        boilerplate = []
        boiler_idx = 0

        if (logs_path / "CITATION.html").exists():
            text = (
                re.compile("<body>(.*?)</body>", re.DOTALL | re.IGNORECASE)
                .findall((logs_path / "CITATION.html").read_text())[0]
                .strip()
            )
            boilerplate.append(
                (boiler_idx, "HTML", f'<div class="boiler-html">{text}</div>')
            )
            boiler_idx += 1

        if (logs_path / "CITATION.md").exists():
            text = (logs_path / "CITATION.md").read_text()
            boilerplate.append((boiler_idx, "Markdown", f"<pre>{text}</pre>\n"))
            boiler_idx += 1

        if (logs_path / "CITATION.tex").exists():
            text = (
                re.compile(
                    r"\\begin{document}(.*?)\\end{document}", re.DOTALL | re.IGNORECASE
                )
                .findall((logs_path / "CITATION.tex").read_text())[0]
                .strip()
            )
            boilerplate.append(
                (
                    boiler_idx,
                    "LaTeX",
                    f"""<pre>{text}</pre>
<h3>Bibliography</h3>
<pre>{Path(pkgrf(self.packagename, 'data/boilerplate.bib')).read_text()}</pre>
""",
                )
            )
            boiler_idx += 1

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchpath=str(self.template_path.parent)),
            trim_blocks=True,
            lstrip_blocks=True,
            autoescape=False,
        )
        report_tpl = env.get_template(self.template_path.name)
        report_render = report_tpl.render(
            sections=self.sections, errors=self.errors, boilerplate=boilerplate
        )

        # Write out report
        self.out_dir.mkdir(parents=True, exist_ok=True)
        (self.out_dir / self.out_filename).write_text(report_render, encoding="UTF-8")
        return len(self.errors)

    @staticmethod
    def _process_orderings(orderings, layout):
        """
        Generate relevant combinations of orderings with observed values.

        Arguments
        ---------
        orderings : :obj:`list` of :obj:`list` of :obj:`str`
            Sections prescribing an ordering to select across sessions, acquisitions, runs, etc.
        layout : :obj:`bids.layout.BIDSLayout`
            The BIDS layout

        Returns
        -------
        entities: :obj:`list` of :obj:`str`
            The relevant orderings that had unique values
        value_combos: :obj:`list` of :obj:`tuple`
            Unique value combinations for the entities

        """
        # get a set of all unique entity combinations
        all_value_combos = {
            tuple(bids_file.get_entities().get(k, None) for k in orderings)
            for bids_file in layout.get()
        }
        # remove the all None member if it exists
        none_member = tuple([None for k in orderings])
        if none_member in all_value_combos:
            all_value_combos.remove(tuple([None for k in orderings]))
        # see what values exist for each entity
        unique_values = [
            {value[idx] for value in all_value_combos} for idx in range(len(orderings))
        ]
        # if all values are None for an entity, we do not want to keep that entity
        keep_idx = [
            False if (len(val_set) == 1 and None in val_set) or not val_set else True
            for val_set in unique_values
        ]
        # the "kept" entities
        entities = list(compress(orderings, keep_idx))
        # the "kept" value combinations
        value_combos = [
            tuple(compress(value_combo, keep_idx)) for value_combo in all_value_combos
        ]
        # sort the value combinations alphabetically from the first entity to the last entity
        value_combos.sort(
            key=lambda entry: tuple(
                str(value) if value is not None else "0" for value in entry
            )
        )

        return entities, value_combos


def run_reports(
    out_dir,
    subject_label,
    run_uuid,
    config=None,
    reportlets_dir=None,
    packagename=None,
):
    """
    Run the reports.

    .. testsetup::

    >>> cwd = os.getcwd()
    >>> os.chdir(tmpdir)

    >>> from pkg_resources import resource_filename
    >>> from shutil import copytree
    >>> test_data_path = resource_filename('niworkflows', 'data/tests/work')
    >>> testdir = Path(tmpdir)
    >>> data_dir = copytree(test_data_path, str(testdir / 'work'))
    >>> (testdir / 'fmriprep').mkdir(parents=True, exist_ok=True)

    .. doctest::

    >>> run_reports(testdir / 'out', '01', 'madeoutuuid', packagename='fmriprep',
    ...             reportlets_dir=testdir / 'work' / 'reportlets')
    0

    .. testcleanup::

    >>> os.chdir(cwd)

    """
    return Report(
        out_dir,
        run_uuid,
        config=config,
        subject_id=subject_label,
        packagename=packagename,
        reportlets_dir=reportlets_dir,
    ).generate_report()


def generate_reports(
    subject_list, output_dir, run_uuid, config=None, work_dir=None, packagename=None
):
    """Execute run_reports on a list of subjects."""
    reportlets_dir = None
    if work_dir is not None:
        reportlets_dir = Path(work_dir) / "reportlets"
    report_errors = [
        run_reports(
            output_dir,
            subject_label,
            run_uuid,
            config=config,
            packagename=packagename,
            reportlets_dir=reportlets_dir,
        )
        for subject_label in subject_list
    ]

    errno = sum(report_errors)
    if errno:
        import logging

        logger = logging.getLogger("cli")
        error_list = ", ".join(
            "%s (%d)" % (subid, err)
            for subid, err in zip(subject_list, report_errors)
            if err
        )
        logger.error(
            "Preprocessing did not finish successfully. Errors occurred while processing "
            "data from participants: %s. Check the HTML reports for details.",
            error_list,
        )
    return errno



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Utilities for tracking and filtering spaces."""
import argparse
import attr
from collections import defaultdict
from itertools import product
from templateflow import api as _tfapi

NONSTANDARD_REFERENCES = [
    "T1w",
    "T2w",
    "anat",
    "fsnative",
    "func",
    "run",
    "sbref",
    "session",
    "individual",
    "dwi",
]
"""List of supported nonstandard reference spaces."""


FSAVERAGE_DENSITY = {
    "fsaverage3": "642",
    "fsaverage4": "2562",
    "fsaverage5": "10k",
    "fsaverage6": "41k",
    "fsaverage": "164k",
}
"""A map of legacy fsaverageX names to surface densities."""

FSAVERAGE_LEGACY = {v: k for k, v in FSAVERAGE_DENSITY.items()}
"""A map of surface densities to legacy fsaverageX names."""


@attr.s(slots=True, frozen=True)
class Reference:
    """
    Represent a (non)standard space specification.

    Examples
    --------
    >>> Reference('MNI152NLin2009cAsym')
    Reference(space='MNI152NLin2009cAsym', spec={})

    >>> Reference('MNI152NLin2009cAsym', {})
    Reference(space='MNI152NLin2009cAsym', spec={})

    >>> Reference('MNI152NLin2009cAsym', None)
    Reference(space='MNI152NLin2009cAsym', spec={})

    >>> Reference('MNI152NLin2009cAsym', {'res': 1})
    Reference(space='MNI152NLin2009cAsym', spec={'res': 1})

    >>> Reference('MNIPediatricAsym', {'cohort': '1'})
    Reference(space='MNIPediatricAsym', spec={'cohort': '1'})

    >>> Reference('func')
    Reference(space='func', spec={})

    >>> # Checks spaces with cohorts:
    >>> Reference('MNIPediatricAsym')
    Traceback (most recent call last):
      ...
    ValueError: standard space "MNIPediatricAsym" is not fully defined.
    ...

    >>> Reference(space='MNI152Lin', spec={'cohort': 1})
    Traceback (most recent call last):
      ...
    ValueError: standard space "MNI152Lin" does not accept ...

    >>> Reference('MNIPediatricAsym', {'cohort': '100'})
    Traceback (most recent call last):
      ...
    ValueError: standard space "MNIPediatricAsym" does not contain ...
    ...

    >>> Reference('MNIPediatricAsym', 'blah')
    Traceback (most recent call last):
      ...
    TypeError: ...

    >>> Reference('shouldraise')
    Traceback (most recent call last):
      ...
    ValueError: space identifier "shouldraise" is invalid.
    ...

    >>> # Check standard property
    >>> Reference('func').standard
    False
    >>> Reference('MNI152Lin').standard
    True
    >>> Reference('MNIPediatricAsym', {'cohort': 1}).standard
    True

    >>> # Equality/inequality checks
    >>> Reference('func') == Reference('func')
    True
    >>> Reference('func') != Reference('MNI152Lin')
    True
    >>> Reference('MNI152Lin', {'res': 1}) == Reference('MNI152Lin', {'res': 1})
    True
    >>> Reference('MNI152Lin', {'res': 1}) == Reference('MNI152Lin', {'res': 2})
    False
    >>> sp1 = Reference('MNIPediatricAsym', {'cohort': 1})
    >>> sp2 = Reference('MNIPediatricAsym', {'cohort': 2})
    >>> sp1 == sp2
    False
    >>> sp1 = Reference('MNIPediatricAsym', {'res': 1, 'cohort': 1})
    >>> sp2 = Reference('MNIPediatricAsym', {'cohort': 1, 'res': 1})
    >>> sp1 == sp2
    True

    """

    _standard_spaces = tuple(_tfapi.templates())

    space = attr.ib(default=None, type=str)
    """Name designating this space."""
    spec = attr.ib(
        factory=dict,
        validator=attr.validators.optional(attr.validators.instance_of(dict)),
    )
    """The dictionary of specs."""
    standard = attr.ib(default=False, repr=False, type=bool)
    """Whether this space is standard or not."""
    dim = attr.ib(default=3, repr=False, type=int)
    """Dimensionality of the sampling manifold."""

    def __attrs_post_init__(self):
        """Extract cohort out of spec."""
        if self.spec is None:
            object.__setattr__(self, "spec", {})

        if self.space.startswith("fsaverage"):
            space = self.space
            object.__setattr__(self, "space", "fsaverage")

            if "den" not in self.spec or space != "fsaverage":
                spec = self.spec.copy()
                spec["den"] = FSAVERAGE_DENSITY[space]
                object.__setattr__(self, "spec", spec)

        if self.space.startswith("fs"):
            object.__setattr__(self, "dim", 2)

        if self.space in self._standard_spaces:
            object.__setattr__(self, "standard", True)

        _cohorts = ["%s" % t for t in _tfapi.TF_LAYOUT.get_cohorts(template=self.space)]
        if "cohort" in self.spec:
            if not _cohorts:
                raise ValueError(
                    'standard space "%s" does not accept a cohort '
                    "specification." % self.space
                )

            if str(self.spec["cohort"]) not in _cohorts:
                raise ValueError(
                    'standard space "%s" does not contain any cohort '
                    'named "%s".' % (self.space, self.spec["cohort"])
                )
        elif _cohorts:
            _cohorts = ", ".join(['"cohort-%s"' % c for c in _cohorts])
            raise ValueError(
                'standard space "%s" is not fully defined.\n'
                "Set a valid cohort selector from: %s." % (self.space, _cohorts)
            )

    @property
    def fullname(self):
        """
        Generate a full-name combining cohort.

        Examples
        --------
        >>> Reference('MNI152Lin').fullname
        'MNI152Lin'

        >>> Reference('MNIPediatricAsym', {'cohort': 1}).fullname
        'MNIPediatricAsym:cohort-1'

        """
        if "cohort" not in self.spec:
            return self.space
        return "%s:cohort-%s" % (self.space, self.spec["cohort"])

    @property
    def legacyname(self):
        """
        Generate a legacy name for fsaverageX spaces.

        Examples
        --------
        >>> Reference(space='fsaverage')
        Reference(space='fsaverage', spec={'den': '164k'})
        >>> Reference(space='fsaverage').legacyname
        'fsaverage'
        >>> Reference(space='fsaverage6')
        Reference(space='fsaverage', spec={'den': '41k'})
        >>> Reference(space='fsaverage6').legacyname
        'fsaverage6'
        >>> # Overwrites density of legacy "fsaverage" specifications
        >>> Reference(space='fsaverage6', spec={'den': '10k'})
        Reference(space='fsaverage', spec={'den': '41k'})
        >>> Reference(space='fsaverage6', spec={'den': '10k'}).legacyname
        'fsaverage6'
        >>> # Return None if no legacy space
        >>> Reference(space='fsaverage', spec={'den': '30k'}).legacyname is None
        True

        """
        if self.space == "fsaverage" and self.spec["den"] in FSAVERAGE_LEGACY:
            return FSAVERAGE_LEGACY[self.spec["den"]]

    @space.validator
    def _check_name(self, attribute, value):
        if value.startswith("fsaverage"):
            return
        valid = list(self._standard_spaces) + NONSTANDARD_REFERENCES
        if value not in valid:
            raise ValueError(
                'space identifier "%s" is invalid.\nValid '
                "identifiers are: %s" % (value, ", ".join(valid))
            )

    def __str__(self):
        """
        Format this reference.

        Examples
        --------
        >>> str(Reference(space='MNIPediatricAsym', spec={'cohort': 2, 'res': 1}))
        'MNIPediatricAsym:cohort-2:res-1'

        """
        return ":".join(
            [self.space] + ["-".join((k, str(v))) for k, v in sorted(self.spec.items())]
        )

    @classmethod
    def from_string(cls, value):
        """
        Parse a string to generate the corresponding list of References.

        .. testsetup::

            >>> if PY_VERSION < (3, 6):
            ...     pytest.skip("This doctest does not work on python <3.6")

        Parameters
        ----------
        value: :obj:`str`
            A string containing a space specification following *fMRIPrep*'s
            language for ``--output-spaces``
            (e.g., ``MNIPediatricAsym:cohort-1:cohort-2:res-1:res-2``).

        Returns
        -------
        spaces : :obj:`list` of :obj:`Reference`
            A list of corresponding spaces given the input string.

        Examples
        --------
        >>> Reference.from_string("MNI152NLin2009cAsym")
        [Reference(space='MNI152NLin2009cAsym', spec={})]

        >>> # Bad space name
        >>> Reference.from_string("shouldraise")
        Traceback (most recent call last):
          ...
        ValueError: space identifier "shouldraise" is invalid.
        ...

        >>> # Missing cohort
        >>> Reference.from_string("MNIPediatricAsym")
        Traceback (most recent call last):
          ...
        ValueError: standard space "MNIPediatricAsym" is not fully defined.
        ...

        >>> Reference.from_string("MNIPediatricAsym:cohort-1")
        [Reference(space='MNIPediatricAsym', spec={'cohort': '1'})]

        >>> Reference.from_string("MNIPediatricAsym:cohort-1:cohort-2")
        [Reference(space='MNIPediatricAsym', spec={'cohort': '1'}),
         Reference(space='MNIPediatricAsym', spec={'cohort': '2'})]

        >>> Reference.from_string("fsaverage:den-10k:den-164k")
        [Reference(space='fsaverage', spec={'den': '10k'}),
         Reference(space='fsaverage', spec={'den': '164k'})]

        >>> Reference.from_string("MNIPediatricAsym:cohort-5:cohort-6:res-2")
        [Reference(space='MNIPediatricAsym', spec={'cohort': '5', 'res': '2'}),
         Reference(space='MNIPediatricAsym', spec={'cohort': '6', 'res': '2'})]

        >>> Reference.from_string("MNIPediatricAsym:cohort-5:cohort-6:res-2:res-iso1.6mm")
        [Reference(space='MNIPediatricAsym', spec={'cohort': '5', 'res': '2'}),
         Reference(space='MNIPediatricAsym', spec={'cohort': '5', 'res': 'iso1.6mm'}),
         Reference(space='MNIPediatricAsym', spec={'cohort': '6', 'res': '2'}),
         Reference(space='MNIPediatricAsym', spec={'cohort': '6', 'res': 'iso1.6mm'})]

        """
        _args = value.split(":")
        spec = defaultdict(list, {})
        for modifier in _args[1:]:
            mitems = modifier.split("-", 1)
            spec[mitems[0]].append(len(mitems) == 1 or mitems[1])

        allspecs = _expand_entities(spec)

        return [cls(_args[0], s) for s in allspecs]


class SpatialReferences:
    """
    Manage specifications of spatial references.

    Examples
    --------
    >>> sp = SpatialReferences([
    ...     'func',
    ...     'fsnative',
    ...     'MNI152NLin2009cAsym',
    ...     'anat',
    ...     'fsaverage5',
    ...     'fsaverage6',
    ...     ('MNIPediatricAsym', {'cohort': '2'}),
    ...     ('MNI152NLin2009cAsym', {'res': 2}),
    ...     ('MNI152NLin2009cAsym', {'res': 1}),
    ... ])
    >>> sp.get_spaces(standard=False)
    ['func', 'fsnative', 'anat']

    >>> sp.get_spaces(standard=False, dim=(3,))
    ['func', 'anat']

    >>> sp.get_spaces(nonstandard=False)
    ['MNI152NLin2009cAsym', 'fsaverage', 'MNIPediatricAsym:cohort-2']

    >>> sp.get_spaces(nonstandard=False, dim=(3,))
    ['MNI152NLin2009cAsym', 'MNIPediatricAsym:cohort-2']

    >>> sp.get_fs_spaces()
    ['fsnative', 'fsaverage5', 'fsaverage6']

    >>> sp.get_standard(full_spec=True)
    [Reference(space='fsaverage', spec={'den': '10k'}),
     Reference(space='fsaverage', spec={'den': '41k'}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 2}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 1})]

    >>> sp.is_cached()
    False
    >>> sp.cached
    Traceback (most recent call last):
     ...
    ValueError: References have not ...

    >>> sp.checkpoint()
    >>> sp.is_cached()
    True
    >>> sp.cached.references
    [Reference(space='func', spec={}),
     Reference(space='fsnative', spec={}),
     Reference(space='MNI152NLin2009cAsym', spec={}),
     Reference(space='anat', spec={}),
     Reference(space='fsaverage', spec={'den': '10k'}),
     Reference(space='fsaverage', spec={'den': '41k'}),
     Reference(space='MNIPediatricAsym', spec={'cohort': '2'}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 2}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 1})]

    >>> sp.cached.get_fs_spaces()
    ['fsnative', 'fsaverage5', 'fsaverage6']

    >>> sp.add(('MNIPediatricAsym', {'cohort': '2'}))
    >>> sp.get_spaces(nonstandard=False, dim=(3,))
    ['MNI152NLin2009cAsym', 'MNIPediatricAsym:cohort-2']

    >>> sp += [('MNIPediatricAsym', {'cohort': '2'})]
    Traceback (most recent call last):
      ...
    ValueError: space ...

    >>> sp += [('MNIPediatricAsym', {'cohort': '1'})]
    >>> sp.get_spaces(nonstandard=False, dim=(3,))
    ['MNI152NLin2009cAsym', 'MNIPediatricAsym:cohort-2', 'MNIPediatricAsym:cohort-1']

    >>> sp.insert(0, ('MNIPediatricAsym', {'cohort': '3'}))
    >>> sp.get_spaces(nonstandard=False, dim=(3,))
    ['MNIPediatricAsym:cohort-3',
     'MNI152NLin2009cAsym',
     'MNIPediatricAsym:cohort-2',
     'MNIPediatricAsym:cohort-1']

    >>> sp.insert(0, ('MNIPediatricAsym', {'cohort': '3'}))
    Traceback (most recent call last):
      ...
    ValueError: space ...

    >>> sp.checkpoint()
    Traceback (most recent call last):
     ...
    ValueError: References have already ...

    >>> sp.checkpoint(force=True)
    >>> sp.cached.references
    [Reference(space='MNIPediatricAsym', spec={'cohort': '3'}),
     Reference(space='func', spec={}),
     Reference(space='fsnative', spec={}),
     Reference(space='MNI152NLin2009cAsym', spec={}),
     Reference(space='anat', spec={}),
     Reference(space='fsaverage', spec={'den': '10k'}),
     Reference(space='fsaverage', spec={'den': '41k'}),
     Reference(space='MNIPediatricAsym', spec={'cohort': '2'}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 2}),
     Reference(space='MNI152NLin2009cAsym', spec={'res': 1}),
     Reference(space='MNIPediatricAsym', spec={'cohort': '1'})]

    """

    __slots__ = ("_refs", "_cached")
    standard_spaces = tuple(_tfapi.templates())
    """List of supported standard reference spaces."""

    @staticmethod
    def check_space(space):
        """Build a :class:`Reference` object."""
        try:
            if isinstance(space, Reference):
                return space
        except IndexError:
            pass

        spec = {}
        if not isinstance(space, str):
            try:
                spec = space[1] or {}
            except IndexError:
                pass
            except TypeError:
                space = (None,)

            space = space[0]
        return Reference(space, spec)

    def __init__(self, spaces=None, checkpoint=False):
        """
        Maintain the bookkeeping of spaces and templates.

        Internal spaces are normalizations required for pipeline execution which
        can vary based on user arguments.
        Output spaces are desired user outputs.
        """
        self._refs = []
        self._cached = None
        if spaces is not None:
            if isinstance(spaces, str):
                spaces = [spaces]
            self.__iadd__(spaces)

            if checkpoint is True:
                self.checkpoint()

    def __iadd__(self, b):
        """Append a list of transforms to the internal list."""
        if not isinstance(b, (list, tuple)):
            raise TypeError("Must be a list.")

        for space in b:
            self.append(space)
        return self

    def __contains__(self, item):
        """Implement the ``in`` builtin."""
        if not self.references:
            return False
        item = self.check_space(item)
        for s in self.references:
            if s == item:
                return True
        return False

    def __str__(self):
        """
        Representation of this object.

        Examples
        --------
        >>> print(SpatialReferences())
        Spatial References: <none>.

        >>> print(SpatialReferences(['MNI152NLin2009cAsym']))
        Spatial References: MNI152NLin2009cAsym

        >>> print(SpatialReferences(['MNI152NLin2009cAsym', 'fsaverage5']))
        Spatial References: MNI152NLin2009cAsym, fsaverage:den-10k

        """
        spaces = ", ".join([str(s) for s in self.references]) or "<none>."
        return "Spatial References: %s" % spaces

    @property
    def references(self):
        """Get all specified references."""
        return self._refs

    @property
    def cached(self):
        """Get cached spaces, raise error if not cached."""
        if not self.is_cached():
            raise ValueError("References have not been cached")
        return self._cached

    def is_cached(self):
        return self._cached is not None

    def checkpoint(self, force=False):
        """Cache and freeze current spaces to separate attribute."""
        if self.is_cached() and not force:
            raise ValueError("References have already been cached")
        self._cached = self.__class__(self.references)

    def add(self, value):
        """Add one more space, without erroring if it exists."""
        if value not in self:
            self._refs += [self.check_space(value)]

    def append(self, value):
        """Concatenate one more space."""
        if value not in self:
            self._refs += [self.check_space(value)]
            return

        raise ValueError('space "%s" already in spaces.' % str(value))

    def insert(self, index, value, error=True):
        """Concatenate one more space."""
        if value not in self:
            self._refs.insert(index, self.check_space(value))
        elif error is True:
            raise ValueError('space "%s" already in spaces.' % str(value))

    def get_spaces(self, standard=True, nonstandard=True, dim=(2, 3)):
        """
        Return space names.

        Parameters
        ----------
        standard : :obj:`bool`, optional
            Return standard spaces.
        nonstandard : :obj:`bool`, optional
            Return nonstandard spaces.
        dim : :obj:`tuple`, optional
            Desired dimensions of the standard spaces (default is ``(2, 3)``)

        Examples
        --------
        >>> spaces = SpatialReferences(['MNI152NLin6Asym', ("fsaverage", {"den": "10k"})])
        >>> spaces.get_spaces()
        ['MNI152NLin6Asym', 'fsaverage']

        >>> spaces.get_spaces(standard=False)
        []

        >>> spaces.get_spaces(dim=(3,))
        ['MNI152NLin6Asym']

        >>> spaces.add(('MNI152NLin6Asym', {'res': '2'}))
        >>> spaces.get_spaces()
        ['MNI152NLin6Asym', 'fsaverage']

        >>> spaces.add(('func', {}))
        >>> spaces.get_spaces()
        ['MNI152NLin6Asym', 'fsaverage', 'func']

        >>> spaces.get_spaces(nonstandard=False)
        ['MNI152NLin6Asym', 'fsaverage']

        >>> spaces.get_spaces(standard=False)
        ['func']

        """
        out = []
        for s in self.references:
            if (
                s.fullname not in out
                and (s.standard is standard or s.standard is not nonstandard)
                and s.dim in dim
            ):
                out.append(s.fullname)
        return out

    def get_standard(self, full_spec=False, dim=(2, 3)):
        """
        Return output spaces.

        Parameters
        ----------
        full_spec : :obj:`bool`
            Return only fully-specified standard references (i.e., they must either
            have density or resolution set).
        dim : :obj:`tuple`, optional
            Desired dimensions of the standard spaces (default is ``(2, 3)``)

        """
        if not full_spec:
            return [s for s in self.references if s.standard and s.dim in dim]

        return [
            s
            for s in self.references
            if s.standard
            and s.dim in dim
            and (hasspec("res", s.spec) or hasspec("den", s.spec))
        ]

    def get_nonstandard(self, full_spec=False, dim=(2, 3)):
        """Return nonstandard spaces."""
        if not full_spec:
            return [s.space for s in self.references if not s.standard and s.dim in dim]
        return [
            s.space
            for s in self.references
            if not s.standard
            and s.dim in dim
            and (hasspec("res", s.spec) or hasspec("den", s.spec))
        ]

    def get_fs_spaces(self):
        """
        Return FreeSurfer spaces.

        Discards nonlegacy fsaverage values (i.e., with nonstandard density value).

        Examples
        --------
        >>> SpatialReferences([
        ...     'fsnative',
        ...     'fsaverage6',
        ...     'fsaverage5',
        ...     'MNI152NLin6Asym',
        ... ]).get_fs_spaces()
        ['fsnative', 'fsaverage6', 'fsaverage5']

        >>> SpatialReferences([
        ...     'fsnative',
        ...     'fsaverage6',
        ...     Reference(space='fsaverage', spec={'den': '30k'})
        ... ]).get_fs_spaces()
        ['fsnative', 'fsaverage6']

        """
        return [
            s.legacyname or s.space
            for s in self.references
            if s.legacyname or s.space == "fsnative"
        ]


class OutputReferencesAction(argparse.Action):
    """Parse spatial references."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Execute parser."""
        spaces = getattr(namespace, self.dest) or SpatialReferences()
        if not values:
            # option was called without any output spaces, so user does not want outputs
            spaces.checkpoint()
        for val in values:
            val = val.rstrip(":")
            if (
                val not in NONSTANDARD_REFERENCES
                and not val.split(":")[0].startswith("fs")
                and ":res-" not in val
                and ":resolution-" not in val
            ):
                # by default, explicitly set volumetric resolution to native
                # relevant discussions:
                # https://github.com/nipreps/niworkflows/pull/457#discussion_r375510227
                # https://github.com/nipreps/niworkflows/pull/494
                val = ":".join((val, "res-native"))
            for sp in Reference.from_string(val):
                spaces.add(sp)
        setattr(namespace, self.dest, spaces)


def hasspec(value, specs):
    """Check whether any of the keys are in a dict."""
    for s in specs:
        if s in value:
            return True
    return False


def format_reference(in_tuple):
    """
    Format a spatial reference given as a tuple.

    Examples
    --------
    >>> format_reference(('MNI152Lin', {'res': 1}))
    'MNI152Lin_res-1'
    >>> format_reference(('MNIPediatricAsym:cohort-2', {'res': 2}))
    'MNIPediatricAsym_cohort-2_res-2'

    """
    out = in_tuple[0].split(":")
    res = in_tuple[1].get("res", None) or in_tuple[1].get("resolution", None)
    if res:
        out.append("-".join(("res", str(res))))
    return "_".join(out)


def reference2dict(in_tuple):
    """
    Split a spatial reference given as a tuple into a dictionary.

    Examples
    --------
    >>> reference2dict(('MNIPediatricAsym:cohort-2', {'res': 2}))
    {'space': 'MNIPediatricAsym', 'cohort': '2', 'resolution': '2'}

    >>> reference2dict(('MNIPediatricAsym:cohort-2', {'res': 2, 'resolution': 1}))
    {'space': 'MNIPediatricAsym', 'cohort': '2', 'resolution': '1'}

    >>> reference2dict(('MNIPediatricAsym:cohort-2', {'res': 2, 'den': '91k'}))
    {'space': 'MNIPediatricAsym', 'cohort': '2', 'resolution': '2', 'density': '91k'}

    """
    tpl_entities = ("space", "cohort")
    retval = {
        tpl_entities[i]: v.split("-")[i] for i, v in enumerate(in_tuple[0].split(":"))
    }
    retval.update(
        {
            "resolution" if k == "res" else "density" if k == "den" else k: f"{v}"
            for k, v in in_tuple[1].items()
        }
    )
    return retval


def _expand_entities(entities):
    """
    Generate multiple replacement queries based on all combinations of values.

    Ported from PyBIDS


    .. testsetup::

        >>> if PY_VERSION < (3, 6):
        ...     pytest.skip("This doctest does not work on python <3.6")

    Examples
    --------
    >>> entities = {'subject': ['01', '02'], 'session': ['1', '2'], 'task': ['rest', 'finger']}
    >>> _expand_entities(entities)
    [{'subject': '01', 'session': '1', 'task': 'rest'},
     {'subject': '01', 'session': '1', 'task': 'finger'},
     {'subject': '01', 'session': '2', 'task': 'rest'},
     {'subject': '01', 'session': '2', 'task': 'finger'},
     {'subject': '02', 'session': '1', 'task': 'rest'},
     {'subject': '02', 'session': '1', 'task': 'finger'},
     {'subject': '02', 'session': '2', 'task': 'rest'},
     {'subject': '02', 'session': '2', 'task': 'finger'}]

    """
    keys = list(entities.keys())
    values = list(product(*[entities[k] for k in keys]))
    return [{k: v for k, v in zip(keys, combs)} for combs in values]



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Miscellaneous utilities."""
import os


__all__ = [
    "get_template_specs",
    "fix_multi_T1w_source_name",
    "add_suffix",
    "read_crashfile",
    "splitext",
    "_copy_any",
    "clean_directory",
]


def get_template_specs(in_template, template_spec=None, default_resolution=1):
    """
    Parse template specifications

    >>> get_template_specs('MNI152NLin2009cAsym', {'suffix': 'T1w'})[1]
    {'resolution': 1}

    >>> get_template_specs('MNI152NLin2009cAsym', {'res': '2', 'suffix': 'T1w'})[1]
    {'resolution': '2'}

    >>> specs = get_template_specs('MNIInfant', {'res': '2', 'cohort': '10', 'suffix': 'T1w'})[1]
    >>> sorted(specs.items())
    [('cohort', '10'), ('resolution', '2')]

    >>> get_template_specs('MNI152NLin2009cAsym',
    ...                    {'suffix': 'T1w', 'cohort': 1})[1] # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    RuntimeError:
    ...

    >>> get_template_specs('MNI152NLin2009cAsym',
    ...                    {'suffix': 'T1w', 'res': '1|2'})[1] # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    RuntimeError:
    ...

    """
    from templateflow.api import get as get_template

    # Massage spec (start creating if None)
    template_spec = template_spec or {}
    template_spec["desc"] = template_spec.get("desc", None)
    template_spec["atlas"] = template_spec.get("atlas", None)
    template_spec["resolution"] = template_spec.pop(
        "res", template_spec.get("resolution", default_resolution)
    )

    common_spec = {"resolution": template_spec["resolution"]}
    if "cohort" in template_spec:
        common_spec["cohort"] = template_spec["cohort"]

    tpl_target_path = get_template(in_template, **template_spec)
    if not tpl_target_path:
        raise RuntimeError(
            """\
Could not find template "{0}" with specs={1}. Please revise your template \
argument.""".format(
                in_template, template_spec
            )
        )

    if isinstance(tpl_target_path, list):
        raise RuntimeError(
            """\
The available template modifiers ({0}) did not select a unique template \
(got "{1}"). Please revise your template argument.""".format(
                template_spec, ", ".join([str(p) for p in tpl_target_path])
            )
        )

    return str(tpl_target_path), common_spec


def fix_multi_T1w_source_name(in_files):
    """
    Make up a generic source name when there are multiple T1s

    >>> fix_multi_T1w_source_name([
    ...     '/path/to/sub-045_ses-test_T1w.nii.gz',
    ...     '/path/to/sub-045_ses-retest_T1w.nii.gz'])
    '/path/to/sub-045_T1w.nii.gz'


    >>> fix_multi_T1w_source_name([
    ...    ('/path/to/sub-045-echo-1_T1w.nii.gz', 'path/to/sub-045-echo-2_T1w.nii.gz')])
    '/path/to/sub-045_T1w.nii.gz'

    """
    import os
    from nipype.utils.filemanip import filename_to_list

    in_file = filename_to_list(in_files)[0]
    if isinstance(in_file, (list, tuple)):
        in_file = in_file[0]

    base, in_file = os.path.split(in_file)
    subject_label = in_file.split("_", 1)[0].split("-")[1]
    return os.path.join(base, "sub-%s_T1w.nii.gz" % subject_label)


def add_suffix(in_files, suffix):
    """
    Wrap nipype's fname_presuffix to conveniently just add a prefix

    >>> add_suffix([
    ...     '/path/to/sub-045_ses-test_T1w.nii.gz',
    ...     '/path/to/sub-045_ses-retest_T1w.nii.gz'], '_test')
    'sub-045_ses-test_T1w_test.nii.gz'

    """
    import os.path as op
    from nipype.utils.filemanip import fname_presuffix, filename_to_list

    return op.basename(fname_presuffix(filename_to_list(in_files)[0], suffix=suffix))


def read_crashfile(path):
    if path.endswith(".pklz"):
        return _read_pkl(path)
    elif path.endswith(".txt"):
        return _read_txt(path)
    raise RuntimeError("unknown crashfile format")


def _read_pkl(path):
    from nipype.utils.filemanip import loadcrash

    crash_data = loadcrash(path)
    data = {"file": path, "traceback": "".join(crash_data["traceback"])}
    if "node" in crash_data:
        data["node"] = crash_data["node"]
        if data["node"].base_dir:
            data["node_dir"] = data["node"].output_dir()
        else:
            data["node_dir"] = "Node crashed before execution"
        data["inputs"] = sorted(data["node"].inputs.trait_get().items())
    return data


def _read_txt(path):
    """Read a txt crashfile

    >>> new_path = Path(__file__).resolve().parent.parent
    >>> test_data_path = new_path / 'data' / 'tests'
    >>> info = _read_txt(test_data_path / 'crashfile.txt')
    >>> info['node']  # doctest: +ELLIPSIS
    '...func_preproc_task_machinegame_run_02_wf.carpetplot_wf.conf_plot'
    >>> info['traceback']  # doctest: +ELLIPSIS
    '...ValueError: zero-size array to reduction operation minimum which has no identity'

    """
    from pathlib import Path

    lines = Path(path).read_text().splitlines()
    data = {"file": str(path)}
    traceback_start = 0
    if lines[0].startswith("Node"):
        data["node"] = lines[0].split(": ", 1)[1].strip()
        data["node_dir"] = lines[1].split(": ", 1)[1].strip()
        inputs = []
        cur_key = ""
        cur_val = ""
        for i, line in enumerate(lines[5:]):
            if not line.strip():
                continue

            if line[0].isspace():
                cur_val += line
                continue

            if cur_val:
                inputs.append((cur_key, cur_val.strip()))

            if line.startswith("Traceback ("):
                traceback_start = i + 5
                break

            cur_key, cur_val = tuple(line.split(" = ", 1))

        data["inputs"] = sorted(inputs)
    else:
        data["node_dir"] = "Node crashed before execution"
    data["traceback"] = "\n".join(lines[traceback_start:]).strip()
    return data


def splitext(fname):
    """
    Split filename in name and extension (.gz safe).

    Examples
    --------
    >>> splitext('some/file.nii.gz')
    ('file', '.nii.gz')
    >>> splitext('some/other/file.nii')
    ('file', '.nii')
    >>> splitext('otherext.tar.gz')
    ('otherext', '.tar.gz')
    >>> splitext('text.txt')
    ('text', '.txt')
    >>> splitext('some/figure.svg')
    ('figure', '.svg')
    >>> splitext('some/figure.svg.gz')
    ('figure', '.svg.gz')
    >>> splitext('some/sub-01_bold.func.gii')
    ('sub-01_bold.func', '.gii')

    """
    from pathlib import Path

    basename = str(Path(fname).name)
    stem = Path(basename.rstrip(".gz")).stem
    return stem, basename[len(stem):]


def _copy_any(src, dst):
    import os
    import gzip
    from shutil import copyfileobj
    from nipype.utils.filemanip import copyfile

    src_isgz = src.endswith(".gz")
    dst_isgz = dst.endswith(".gz")
    if not src_isgz and not dst_isgz:
        copyfile(src, dst, copy=True, use_hardlink=True)
        return False  # Make sure we do not reuse the hardlink later

    # Unlink target (should not exist)
    if os.path.exists(dst):
        os.unlink(dst)

    src_open = gzip.open if src_isgz else open
    with src_open(src, "rb") as f_in:
        with open(dst, "wb") as f_out:
            if dst_isgz:
                # Remove FNAME header from gzip (nipreps/fmriprep#1480)
                gz_out = gzip.GzipFile("", "wb", 9, f_out, 0.0)
                copyfileobj(f_in, gz_out)
                gz_out.close()
            else:
                copyfileobj(f_in, f_out)

    return True


def clean_directory(path):
    """
    Clears a directory of all contents.

    Returns `True` if no content remains. If any content cannot be removed, returns `False`.

    Notes
    -----
    This function is not guaranteed to work across multiple threads or processes.

    """
    from pathlib import Path
    import shutil

    try:
        for f in Path(path).iterdir():
            if f.is_file() or f.is_symlink():
                f.unlink()
            elif f.is_dir():
                shutil.rmtree(str(f))
    except OSError:
        return False
    return True


def pass_dummy_scans(algo_dummy_scans, dummy_scans=None):
    """
    Graft manually provided number of dummy scans, if necessary.

    Parameters
    ----------
    algo_dummy_scans : int
        number of volumes to skip determined by an algorithm
    dummy_scans : int or None
        number of volumes to skip determined by the user

    Returns
    -------
    skip_vols_num : int
        number of volumes to skip

    """
    if dummy_scans is None:
        return algo_dummy_scans
    return dummy_scans


def check_valid_fs_license():
    """
    Run ``mri_convert`` to assess FreeSurfer access to a license.

    Returns
    -------
    valid : :obj:`bool`
        FreeSurfer successfully executed (valid license)

    """
    from pathlib import Path
    import subprocess as sp
    from tempfile import TemporaryDirectory
    from pkg_resources import resource_filename

    with TemporaryDirectory() as tmpdir:
        # quick FreeSurfer command
        _cmd = (
            "mri_convert",
            resource_filename("niworkflows", "data/sentinel.nii.gz"),
            str(Path(tmpdir) / "out.mgz"),
        )
        proc = sp.run(_cmd, stdout=sp.PIPE, stderr=sp.STDOUT)
    return proc.returncode == 0 and "ERROR:" not in proc.stdout.decode()


def unlink(pathlike, missing_ok=False):
    """Backport of Path.unlink from Python 3.8+ with missing_ok keyword"""
    # PY37 hack; drop when python_requires >= 3.8
    try:
        os.unlink(pathlike)
    except FileNotFoundError:
        if not missing_ok:
            raise


if __name__ == "__main__":
    pass



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Handling surfaces."""
import os
import re
from pathlib import Path
from collections import defaultdict

import numpy as np
import nibabel as nb

from nipype.utils.filemanip import fname_presuffix
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    TraitedSpec,
    DynamicTraitedSpec,
    SimpleInterface,
    CommandLine,
    CommandLineInputSpec,
    File,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
    Undefined,
)


SECONDARY_ANAT_STRUC = {
    "smoothwm": "GrayWhite",
    "pial": "Pial",
    "midthickness": "GrayMid",
}


class _NormalizeSurfInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True, desc="Freesurfer-generated GIFTI file")
    transform_file = File(exists=True, desc="FSL or LTA affine transform file")


class _NormalizeSurfOutputSpec(TraitedSpec):
    out_file = File(desc="output file with re-centered GIFTI coordinates")


class NormalizeSurf(SimpleInterface):
    """
    Normalize a FreeSurfer-generated GIFTI image.

    FreeSurfer includes an offset to the center of the brain volume that is not
    respected by all software packages.
    Normalization involves adding this offset to the coordinates of all
    vertices, and zeroing out that offset, to ensure consistent behavior
    across software packages.
    In particular, this normalization is consistent with the Human Connectome
    Project pipeline (see `AlgorithmSurfaceApplyAffine`_ and
    `FreeSurfer2CaretConvertAndRegisterNonlinear`_), although the the HCP
    may not zero out the offset.

    GIFTI files with ``midthickness``/``graymid`` in the name are also updated
    to include the following metadata entries::

        {
            AnatomicalStructureSecondary: MidThickness,
            GeometricType: Anatomical
        }

    This interface is intended to be applied uniformly to GIFTI surface files
    generated from the ``?h.white``/``?h.smoothwm`` and ``?h.pial`` surfaces,
    as well as externally-generated ``?h.midthickness``/``?h.graymid`` files.
    In principle, this should apply safely to any other surface, although it is
    less relevant to surfaces that don't describe an anatomical structure.

    .. _AlgorithmSurfaceApplyAffine: https://github.com/Washington-University/workbench\
/blob/1b79e56/src/Algorithms/AlgorithmSurfaceApplyAffine.cxx#L73-L91

    .. _FreeSurfer2CaretConvertAndRegisterNonlinear: https://github.com/Washington-University/\
Pipelines/blob/ae69b9a/PostFreeSurfer/scripts/FreeSurfer2CaretConvertAndRegisterNonlinear.sh\
#L147-154

    """

    input_spec = _NormalizeSurfInputSpec
    output_spec = _NormalizeSurfOutputSpec

    def _run_interface(self, runtime):
        transform_file = self.inputs.transform_file
        if not isdefined(transform_file):
            transform_file = None
        self._results["out_file"] = normalize_surfs(
            self.inputs.in_file, transform_file, newpath=runtime.cwd
        )
        return runtime


class _Path2BIDSInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, desc="input GIFTI file")


class _Path2BIDSOutputSpec(DynamicTraitedSpec):
    extension = traits.Str()


class Path2BIDS(SimpleInterface):
    """
    Extract BIDS entities from paths using a pattern.

    Default pattern is given for Gifti surfaces.

    >>> Path2BIDS(in_file='_fix_surfs0/rh.pial.surf.gii').run().outputs
    <BLANKLINE>
    extension = .surf.gii
    hemi = R
    suffix = pial
    <BLANKLINE>

    >>> Path2BIDS(in_file='_fix_surfs0/rh.pial.gii').run().outputs
    <BLANKLINE>
    extension = .gii
    hemi = R
    suffix = pial
    <BLANKLINE>

    >>> Path2BIDS(in_file='_fix_surfs0/rh.smoothwm_converted.gii').run().outputs
    <BLANKLINE>
    extension = .gii
    hemi = R
    suffix = smoothwm
    <BLANKLINE>

    >>> Path2BIDS(in_file='_fix_surfs0/rh.smoothwm_converted.func.gii').run().outputs
    <BLANKLINE>
    extension = .func.gii
    hemi = R
    suffix = smoothwm
    <BLANKLINE>

    """

    input_spec = _Path2BIDSInputSpec
    output_spec = _Path2BIDSOutputSpec
    _pattern = re.compile(
        r"(?P<hemi>[lr])h.(?P<suffix>(wm|smoothwm|pial|midthickness|"
        r"inflated|vinflated|sphere|flat))[\w\d_-]*(?P<extprefix>\.\w+)?"
    )
    _excluded = ("extprefix",)

    def __init__(self, pattern=None, **inputs):
        """Initialize the interface."""
        super().__init__(**inputs)
        if pattern:
            self._pattern = re.compile(pattern)

    def _outputs(self):
        outputs = self.output_spec()
        outputs.trait_set(
            trait_change_notify=False,
            **{
                entity: Undefined
                for entity in self._pattern.groupindex.keys()
                if entity not in self._excluded
            },
        )
        return outputs

    def _run_interface(self, runtime):
        in_file = Path(self.inputs.in_file)
        extension = "".join(in_file.suffixes[-((in_file.suffixes[-1] == ".gz") + 1):])
        info = self._pattern.match(in_file.name[: -len(extension)]).groupdict()
        self._results["extension"] = f"{info.pop('extprefix', None) or ''}{extension}"
        self._results.update(info)
        if "hemi" in self._results:
            self._results["hemi"] = self._results["hemi"].upper()
        return runtime


class _GiftiNameSourceInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True, desc="input GIFTI file")
    pattern = traits.Str(
        mandatory=True, desc='input file name pattern (must capture named group "LR")'
    )
    template = traits.Str(mandatory=True, desc="output file name template")
    template_kwargs = traits.Dict(desc="additional template keyword value pairs")


class _GiftiNameSourceOutputSpec(TraitedSpec):
    out_name = traits.Str(desc="(partial) filename formatted according to template")


class GiftiNameSource(SimpleInterface):
    r"""
    Construct a new filename for a GIFTI file.

    Construct a new filename based on an input filename, a matching pattern,
    and a related template, with optionally additional keywords.

    This interface is intended for use with GIFTI files, to generate names
    conforming to Section 9.0 of the `GIFTI Standard`_.

    Patterns are expected to have named groups, including one named "LR" that
    matches "l" or "r".
    These groups must correspond to named format elements in the template.

    .. testsetup::

       >>> open('lh.pial.gii', 'w').close()
       >>> open('rh.fsaverage.gii', 'w').close()

    Examples
    --------
    >>> surf_namer = GiftiNameSource()
    >>> surf_namer.inputs.pattern = r'(?P<LR>[lr])h.(?P<surf>\w+).gii'
    >>> surf_namer.inputs.template = r'{surf}.{LR}.surf'
    >>> surf_namer.inputs.in_file = 'lh.pial.gii'
    >>> res = surf_namer.run()
    >>> res.outputs.out_name
    'pial.L.surf'

    >>> func_namer = GiftiNameSource()
    >>> func_namer.inputs.pattern = r'(?P<LR>[lr])h.(?P<space>\w+).gii'
    >>> func_namer.inputs.template = r'space-{space}.{LR}.func'
    >>> func_namer.inputs.in_file = 'rh.fsaverage.gii'
    >>> res = func_namer.run()
    >>> res.outputs.out_name
    'space-fsaverage.R.func'

    >>> namer = GiftiNameSource()
    >>> namer.inputs.pattern = r'(?P<LR>[lr])h.(?P<space>\w+).gii'
    >>> namer.inputs.template = r'space-{space}_density-{density}_hemi-{LR}.func'
    >>> namer.inputs.in_file = 'rh.fsaverage.gii'
    >>> namer.inputs.template_kwargs = {'density': '10k'}
    >>> res = namer.run()
    >>> res.outputs.out_name
    'space-fsaverage_density-10k_hemi-R.func'

    .. testcleanup::

    >>> import os
    >>> os.unlink('lh.pial.gii')
    >>> os.unlink('rh.fsaverage.gii')

    .. _GIFTI Standard: https://www.nitrc.org/frs/download.php/2871/GIFTI_Surface_Format.pdf
    """
    input_spec = _GiftiNameSourceInputSpec
    output_spec = _GiftiNameSourceOutputSpec

    def _run_interface(self, runtime):
        in_format = re.compile(self.inputs.pattern)
        in_file = os.path.basename(self.inputs.in_file)
        info = in_format.match(in_file).groupdict()
        info["LR"] = info["LR"].upper()
        if self.inputs.template_kwargs:
            info.update(self.inputs.template_kwargs)
        filefmt = self.inputs.template
        self._results["out_name"] = filefmt.format(**info)
        return runtime


class _GiftiSetAnatomicalStructureInputSpec(BaseInterfaceInputSpec):
    in_file = File(
        mandatory=True, exists=True, desc='GIFTI file beginning with "lh." or "rh."'
    )


class _GiftiSetAnatomicalStructureOutputSpec(TraitedSpec):
    out_file = File(desc="output file with updated AnatomicalStructurePrimary entry")


class GiftiSetAnatomicalStructure(SimpleInterface):
    """
    Set AnatomicalStructurePrimary attribute of GIFTI image based on filename.

    For files that begin with ``lh.`` or ``rh.``, update the metadata to
    include::

        {
            AnatomicalStructurePrimary: (CortexLeft | CortexRight),
        }

    If ``AnatomicalStructurePrimary`` is already set, this function has no
    effect.

    """

    input_spec = _GiftiSetAnatomicalStructureInputSpec
    output_spec = _GiftiSetAnatomicalStructureOutputSpec

    def _run_interface(self, runtime):
        img = nb.load(self.inputs.in_file)
        if any(nvpair.name == "AnatomicalStruturePrimary" for nvpair in img.meta.data):
            out_file = self.inputs.in_file
        else:
            fname = os.path.basename(self.inputs.in_file)
            if fname[:3] in ("lh.", "rh."):
                asp = "CortexLeft" if fname[0] == "l" else "CortexRight"
            else:
                raise ValueError(
                    "AnatomicalStructurePrimary cannot be derived from filename"
                )
            img.meta.data.insert(
                0, nb.gifti.GiftiNVPairs("AnatomicalStructurePrimary", asp)
            )
            out_file = os.path.join(runtime.cwd, fname)
            img.to_filename(out_file)
        self._results["out_file"] = out_file
        return runtime


class _GiftiToCSVInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True, desc="GIFTI file")
    itk_lps = traits.Bool(False, usedefault=True, desc="flip XY axes")


class _GiftiToCSVOutputSpec(TraitedSpec):
    out_file = File(desc="output csv file")


class GiftiToCSV(SimpleInterface):
    """Converts GIfTI files to CSV to make them ammenable to use with
    ``antsApplyTransformsToPoints``."""

    input_spec = _GiftiToCSVInputSpec
    output_spec = _GiftiToCSVOutputSpec

    def _run_interface(self, runtime):
        gii = nb.load(self.inputs.in_file)
        data = gii.darrays[0].data

        if self.inputs.itk_lps:  # ITK: flip X and Y around 0
            data[:, :2] *= -1

        # antsApplyTransformsToPoints requires 5 cols with headers
        csvdata = np.hstack((data, np.zeros((data.shape[0], 3))))

        out_file = fname_presuffix(
            self.inputs.in_file, newpath=runtime.cwd, use_ext=False, suffix="points.csv"
        )
        np.savetxt(
            out_file,
            csvdata,
            delimiter=",",
            header="x,y,z,t,label,comment",
            fmt=["%.5f"] * 4 + ["%d"] * 2,
        )
        self._results["out_file"] = out_file
        return runtime


class _CSVToGiftiInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True, desc="CSV file")
    gii_file = File(mandatory=True, exists=True, desc="reference GIfTI file")
    itk_lps = traits.Bool(False, usedefault=True, desc="flip XY axes")


class _CSVToGiftiOutputSpec(TraitedSpec):
    out_file = File(desc="output GIfTI file")


class CSVToGifti(SimpleInterface):
    """Converts CSV files back to GIfTI, after moving vertices with
    ``antsApplyTransformToPoints``."""

    input_spec = _CSVToGiftiInputSpec
    output_spec = _CSVToGiftiOutputSpec

    def _run_interface(self, runtime):
        gii = nb.load(self.inputs.gii_file)
        data = np.loadtxt(
            self.inputs.in_file, delimiter=",", skiprows=1, usecols=(0, 1, 2)
        )

        if self.inputs.itk_lps:  # ITK: flip X and Y around 0
            data[:, :2] *= -1

        gii.darrays[0].data = data[:, :3].astype(gii.darrays[0].data.dtype)
        out_file = fname_presuffix(
            self.inputs.gii_file, newpath=runtime.cwd, suffix=".transformed"
        )
        gii.to_filename(out_file)
        self._results["out_file"] = out_file
        return runtime


class _SurfacesToPointCloudInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiPath(
        File(exists=True), mandatory=True, desc="input GIfTI files"
    )
    out_file = File("pointcloud.ply", usedefault=True, desc="output file name")


class _SurfacesToPointCloudOutputSpec(TraitedSpec):
    out_file = File(desc="output pointcloud in PLY format")


class SurfacesToPointCloud(SimpleInterface):
    """Converts multiple surfaces into a pointcloud with corresponding normals
    to then apply Poisson reconstruction"""

    input_spec = _SurfacesToPointCloudInputSpec
    output_spec = _SurfacesToPointCloudOutputSpec

    def _run_interface(self, runtime):
        from pathlib import Path

        giis = [nb.load(g) for g in self.inputs.in_files]
        vertices = np.vstack([g.darrays[0].data for g in giis])
        norms = np.vstack(
            [vertex_normals(g.darrays[0].data, g.darrays[1].data) for g in giis]
        )
        out_file = Path(self.inputs.out_file).resolve()
        pointcloud2ply(vertices, norms, out_file=out_file)
        self._results["out_file"] = str(out_file)
        return runtime


class _PoissonReconInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="--in %s",
        desc="input PLY pointcloud (vertices + normals)",
    )
    out_file = File(
        argstr="--out %s",
        keep_extension=True,
        name_source=["in_file"],
        name_template="%s_avg",
        desc="output PLY triangular mesh",
    )


class _PoissonReconOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output PLY triangular mesh")


class PoissonRecon(CommandLine):
    """Runs Poisson Reconstruction on a cloud of points + normals
    given in PLY format.
    See https://github.com/mkazhdan/PoissonRecon
    """

    input_spec = _PoissonReconInputSpec
    output_spec = _PoissonReconOutputSpec
    _cmd = "PoissonRecon"


class _PLYtoGiftiInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="input PLY file")
    surf_key = traits.Str(mandatory=True, desc="reference GIfTI file")


class _PLYtoGiftiOutputSpec(TraitedSpec):
    out_file = File(desc="output GIfTI file")


class PLYtoGifti(SimpleInterface):
    """Convert surfaces from PLY to GIfTI"""

    input_spec = _PLYtoGiftiInputSpec
    output_spec = _PLYtoGiftiOutputSpec

    def _run_interface(self, runtime):
        from pathlib import Path

        meta = {
            "GeometricType": "Anatomical",
            "VolGeomWidth": "256",
            "VolGeomHeight": "256",
            "VolGeomDepth": "256",
            "VolGeomXsize": "1.0",
            "VolGeomYsize": "1.0",
            "VolGeomZsize": "1.0",
            "VolGeomX_R": "-1.0",
            "VolGeomX_A": "0.0",
            "VolGeomX_S": "0.0",
            "VolGeomY_R": "0.0",
            "VolGeomY_A": "0.0",
            "VolGeomY_S": "-1.0",
            "VolGeomZ_R": "0.0",
            "VolGeomZ_A": "1.0",
            "VolGeomZ_S": "0.0",
            "VolGeomC_R": "0.0",
            "VolGeomC_A": "0.0",
            "VolGeomC_S": "0.0",
        }
        meta["AnatomicalStructurePrimary"] = "Cortex%s" % (
            "Left" if self.inputs.surf_key.startswith("lh") else "Right"
        )
        meta["AnatomicalStructureSecondary"] = SECONDARY_ANAT_STRUC[
            self.inputs.surf_key.split(".")[-1]
        ]
        meta["Name"] = "%s_average.gii" % self.inputs.surf_key

        out_file = Path(runtime.cwd) / meta["Name"]
        out_file = ply2gii(self.inputs.in_file, meta, out_file=out_file)
        self._results["out_file"] = str(out_file)
        return runtime


class _UnzipJoinedSurfacesInputSpec(BaseInterfaceInputSpec):
    in_files = traits.List(
        InputMultiPath(File(exists=True), mandatory=True, desc="input GIfTI files")
    )


class _UnzipJoinedSurfacesOutputSpec(TraitedSpec):
    out_files = traits.List(
        OutputMultiPath(File(exists=True), desc="output pointcloud in PLY format")
    )
    surf_keys = traits.List(traits.Str, desc="surface identifier keys")


class UnzipJoinedSurfaces(SimpleInterface):
    """Unpack surfaces by identifier keys"""

    input_spec = _UnzipJoinedSurfacesInputSpec
    output_spec = _UnzipJoinedSurfacesOutputSpec

    def _run_interface(self, runtime):
        from pathlib import Path

        groups = defaultdict(list)
        in_files = [it for items in self.inputs.in_files for it in items]

        for f in in_files:
            bname = Path(f).name
            groups[bname.split("_")[0]].append(f)

        self._results["out_files"] = [sorted(els) for els in groups.values()]
        self._results["surf_keys"] = list(groups.keys())

        return runtime


def normalize_surfs(in_file, transform_file, newpath=None):
    """
    Re-center GIFTI coordinates to fit align to native T1w space.

    For midthickness surfaces, add MidThickness metadata

    Coordinate update based on:
    https://github.com/Washington-University/workbench/blob/1b79e56/src/Algorithms/AlgorithmSurfaceApplyAffine.cxx#L73-L91
    and
    https://github.com/Washington-University/Pipelines/blob/ae69b9a/PostFreeSurfer/scripts/FreeSurfer2CaretConvertAndRegisterNonlinear.sh#L147
    """

    img = nb.load(in_file)
    transform = load_transform(transform_file)
    pointset = img.get_arrays_from_intent("NIFTI_INTENT_POINTSET")[0]
    coords = pointset.data.T
    c_ras_keys = ("VolGeomC_R", "VolGeomC_A", "VolGeomC_S")
    ras = np.array([[float(pointset.metadata[key])] for key in c_ras_keys])
    ones = np.ones((1, coords.shape[1]), dtype=coords.dtype)
    # Apply C_RAS translation to coordinates, then transform
    pointset.data = transform.dot(np.vstack((coords + ras, ones)))[:3].T.astype(
        coords.dtype
    )

    secondary = nb.gifti.GiftiNVPairs("AnatomicalStructureSecondary", "MidThickness")
    geom_type = nb.gifti.GiftiNVPairs("GeometricType", "Anatomical")
    has_ass = has_geo = False
    for nvpair in pointset.meta.data:
        # Remove C_RAS translation from metadata to avoid double-dipping in FreeSurfer
        if nvpair.name in c_ras_keys:
            nvpair.value = "0.000000"
        # Check for missing metadata
        elif nvpair.name == secondary.name:
            has_ass = True
        elif nvpair.name == geom_type.name:
            has_geo = True
    fname = os.path.basename(in_file)
    # Update metadata for MidThickness/graymid surfaces
    if "midthickness" in fname.lower() or "graymid" in fname.lower():
        if not has_ass:
            pointset.meta.data.insert(1, secondary)
        if not has_geo:
            pointset.meta.data.insert(2, geom_type)

    if newpath is not None:
        newpath = os.getcwd()
    out_file = os.path.join(newpath, fname)
    img.to_filename(out_file)
    return out_file


def load_transform(fname):
    """Load affine transform from file

    Parameters
    ----------
    fname : str or None
        Filename of an LTA or FSL-style MAT transform file.
        If ``None``, return an identity transform

    Returns
    -------
    affine : (4, 4) numpy.ndarray
    """
    if fname is None:
        return np.eye(4)

    if fname.endswith(".mat"):
        return np.loadtxt(fname)
    elif fname.endswith(".lta"):
        with open(fname, "rb") as fobj:
            for line in fobj:
                if line.startswith(b"1 4 4"):
                    break
            lines = fobj.readlines()[:4]
        return np.genfromtxt(lines)

    raise ValueError("Unknown transform type; pass FSL (.mat) or LTA (.lta)")


def vertex_normals(vertices, faces):
    """Calculates the normals of a triangular mesh"""

    def normalize_v3(arr):
        """ Normalize a numpy array of 3 component vectors shape=(n,3) """
        lens = np.sqrt(arr[:, 0] ** 2 + arr[:, 1] ** 2 + arr[:, 2] ** 2)
        arr /= lens[:, np.newaxis]

    tris = vertices[faces]
    facenorms = np.cross(tris[::, 1] - tris[::, 0], tris[::, 2] - tris[::, 0])
    normalize_v3(facenorms)

    norm = np.zeros(vertices.shape, dtype=vertices.dtype)
    norm[faces[:, 0]] += facenorms
    norm[faces[:, 1]] += facenorms
    norm[faces[:, 2]] += facenorms
    normalize_v3(norm)
    return norm


def pointcloud2ply(vertices, normals, out_file=None):
    """Converts the file to PLY format"""
    from pathlib import Path
    import pandas as pd
    from pyntcloud import PyntCloud

    df = pd.DataFrame(np.hstack((vertices, normals)))
    df.columns = ["x", "y", "z", "nx", "ny", "nz"]
    cloud = PyntCloud(df)

    if out_file is None:
        out_file = Path("pointcloud.ply").resolve()

    cloud.to_file(str(out_file))
    return out_file


def ply2gii(in_file, metadata, out_file=None):
    """Convert from ply to GIfTI"""
    from pathlib import Path
    from numpy import eye
    from nibabel.gifti import (
        GiftiMetaData,
        GiftiCoordSystem,
        GiftiImage,
        GiftiDataArray,
    )
    from pyntcloud import PyntCloud

    in_file = Path(in_file)
    surf = PyntCloud.from_file(str(in_file))

    # Update centroid metadata
    metadata.update(
        zip(
            ("SurfaceCenterX", "SurfaceCenterY", "SurfaceCenterZ"),
            ["%.4f" % c for c in surf.centroid],
        )
    )

    # Prepare data arrays
    da = (
        GiftiDataArray(
            data=surf.xyz.astype("float32"),
            datatype="NIFTI_TYPE_FLOAT32",
            intent="NIFTI_INTENT_POINTSET",
            meta=GiftiMetaData.from_dict(metadata),
            coordsys=GiftiCoordSystem(xform=eye(4), xformspace=3),
        ),
        GiftiDataArray(
            data=surf.mesh.values,
            datatype="NIFTI_TYPE_INT32",
            intent="NIFTI_INTENT_TRIANGLE",
            coordsys=None,
        ),
    )
    surfgii = GiftiImage(darrays=da)

    if out_file is None:
        out_file = fname_presuffix(
            in_file.name, suffix=".gii", use_ext=False, newpath=str(Path.cwd())
        )

    surfgii.to_filename(str(out_file))
    return out_file


def get_gii_meta(in_file):
    from nibabel import load

    if isinstance(in_file, list):
        in_file = in_file[0]
    gii = load(in_file)
    return gii.darrays[0].meta.metadata



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""NeuroImaging Workflows (NIWorkflows) is a selection of image processing workflows."""
import logging

from .__about__ import __version__, __packagename__, __copyright__, __credits__


__all__ = [
    "__version__",
    "__packagename__",
    "__copyright__",
    "__credits__",
    "NIWORKFLOWS_LOG",
]

NIWORKFLOWS_LOG = logging.getLogger(__packagename__)
NIWORKFLOWS_LOG.setLevel(logging.INFO)

try:
    import matplotlib

    matplotlib.use("Agg")
except ImportError:
    pass



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Nibabel-based interfaces."""
import numpy as np
import nibabel as nb
from nipype import logging
from nipype.utils.filemanip import fname_presuffix
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterfaceInputSpec,
    File,
    SimpleInterface,
    OutputMultiObject,
    InputMultiObject,
)

IFLOGGER = logging.getLogger("nipype.interface")


class _ApplyMaskInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="an image")
    in_mask = File(exists=True, mandatory=True, desc="a mask")
    threshold = traits.Float(
        0.5, usedefault=True, desc="a threshold to the mask, if it is nonbinary"
    )


class _ApplyMaskOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="masked file")


class ApplyMask(SimpleInterface):
    """Mask the input given a mask."""

    input_spec = _ApplyMaskInputSpec
    output_spec = _ApplyMaskOutputSpec

    def _run_interface(self, runtime):
        img = nb.load(self.inputs.in_file)
        msknii = nb.load(self.inputs.in_mask)
        msk = msknii.get_fdata() > self.inputs.threshold

        self._results["out_file"] = fname_presuffix(
            self.inputs.in_file, suffix="_masked", newpath=runtime.cwd
        )

        if img.dataobj.shape[:3] != msk.shape:
            raise ValueError("Image and mask sizes do not match.")

        if not np.allclose(img.affine, msknii.affine):
            raise ValueError("Image and mask affines are not similar enough.")

        if img.dataobj.ndim == msk.ndim + 1:
            msk = msk[..., np.newaxis]

        masked = img.__class__(img.dataobj * msk, None, img.header)
        masked.to_filename(self._results["out_file"])
        return runtime


class _BinarizeInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="input image")
    thresh_low = traits.Float(mandatory=True, desc="non-inclusive lower threshold")


class _BinarizeOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="masked file")
    out_mask = File(exists=True, desc="output mask")


class Binarize(SimpleInterface):
    """Binarizes the input image applying the given thresholds."""

    input_spec = _BinarizeInputSpec
    output_spec = _BinarizeOutputSpec

    def _run_interface(self, runtime):
        img = nb.load(self.inputs.in_file)

        self._results["out_file"] = fname_presuffix(
            self.inputs.in_file, suffix="_masked", newpath=runtime.cwd
        )
        self._results["out_mask"] = fname_presuffix(
            self.inputs.in_file, suffix="_mask", newpath=runtime.cwd
        )

        data = img.get_fdata()
        mask = data > self.inputs.thresh_low
        data[~mask] = 0.0
        masked = img.__class__(data, img.affine, img.header)
        masked.to_filename(self._results["out_file"])

        img.header.set_data_dtype("uint8")
        maskimg = img.__class__(mask.astype("uint8"), img.affine, img.header)
        maskimg.to_filename(self._results["out_mask"])

        return runtime


class _SplitSeriesInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="input 4d image")


class _SplitSeriesOutputSpec(TraitedSpec):
    out_files = OutputMultiObject(File(exists=True), desc="output list of 3d images")


class SplitSeries(SimpleInterface):
    """Split a 4D dataset along the last dimension into a series of 3D volumes."""

    input_spec = _SplitSeriesInputSpec
    output_spec = _SplitSeriesOutputSpec

    def _run_interface(self, runtime):
        in_file = self.inputs.in_file
        img = nb.load(in_file)
        extra_dims = tuple(dim for dim in img.shape[3:] if dim > 1) or (1,)
        if len(extra_dims) != 1:
            raise ValueError(f"Invalid shape {'x'.join(str(s) for s in img.shape)}")
        img = img.__class__(
            img.dataobj.reshape(img.shape[:3] + extra_dims), img.affine, img.header
        )

        self._results["out_files"] = []
        for i, img_3d in enumerate(nb.four_to_three(img)):
            out_file = fname_presuffix(
                in_file, suffix=f"_idx-{i:03}", newpath=runtime.cwd
            )
            img_3d.to_filename(out_file)
            self._results["out_files"].append(out_file)

        return runtime


class _MergeSeriesInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiObject(
        File(exists=True, mandatory=True, desc="input list of 3d images")
    )
    allow_4D = traits.Bool(
        True, usedefault=True, desc="whether 4D images are allowed to be concatenated"
    )
    affine_tolerance = traits.Float(desc="Absolute tolerance allowed between image affines")


class _MergeSeriesOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="output 4d image")


class MergeSeries(SimpleInterface):
    """Merge a series of 3D volumes along the last dimension into a single 4D image."""

    input_spec = _MergeSeriesInputSpec
    output_spec = _MergeSeriesOutputSpec

    def _run_interface(self, runtime):
        nii_list = []
        aff0 = None
        for f in self.inputs.in_files:
            filenii = nb.squeeze_image(nb.load(f))
            if self.inputs.affine_tolerance:
                if aff0 is None:
                    aff0 = filenii.affine
                elif not np.allclose(aff0, filenii.affine, atol=self.inputs.affine_tolerance):
                    raise ValueError(
                        "Difference in affines greater than allowed tolerance "
                        f"{self.inputs.affine_tolerance}"
                    )
            ndim = filenii.dataobj.ndim
            if ndim == 3:
                nii_list.append(filenii)
                continue
            elif self.inputs.allow_4D and ndim == 4:
                nii_list += nb.four_to_three(filenii)
                continue
            else:
                raise ValueError(
                    "Input image has an incorrect number of dimensions" f" ({ndim})."
                )

        img_4d = nb.concat_images(
            nii_list,
            check_affines=not bool(self.inputs.affine_tolerance)
        )
        out_file = fname_presuffix(
            self.inputs.in_files[0], suffix="_merged", newpath=runtime.cwd
        )
        img_4d.to_filename(out_file)

        self._results["out_file"] = out_file
        return runtime


class _RegridToZoomsInputSpec(BaseInterfaceInputSpec):
    in_file = File(
        exists=True, mandatory=True, desc="a file whose resolution is to change"
    )
    zooms = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        mandatory=True,
        desc="the new resolution",
    )
    order = traits.Int(3, usedefault=True, desc="order of interpolator")
    clip = traits.Bool(
        True,
        usedefault=True,
        desc="clip the data array within the original image's range",
    )
    smooth = traits.Either(
        traits.Bool(),
        traits.Float(),
        default=False,
        usedefault=True,
        desc="apply gaussian smoothing before resampling",
    )


class _RegridToZoomsOutputSpec(TraitedSpec):
    out_file = File(exists=True, dec="the regridded file")


class RegridToZooms(SimpleInterface):
    """Change the resolution of an image (regrid)."""

    input_spec = _RegridToZoomsInputSpec
    output_spec = _RegridToZoomsOutputSpec

    def _run_interface(self, runtime):
        from ..utils.images import resample_by_spacing

        self._results["out_file"] = fname_presuffix(
            self.inputs.in_file, suffix="_regrid", newpath=runtime.cwd
        )
        resample_by_spacing(
            self.inputs.in_file,
            self.inputs.zooms,
            order=self.inputs.order,
            clip=self.inputs.clip,
            smooth=self.inputs.smooth,
        ).to_filename(self._results["out_file"])
        return runtime


class _DemeanImageInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="image to be demeaned")
    in_mask = File(
        exists=True, mandatory=True, desc="mask where median will be calculated"
    )
    only_mask = traits.Bool(False, usedefault=True, desc="demean only within mask")


class _DemeanImageOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="demeaned image")


class DemeanImage(SimpleInterface):
    input_spec = _DemeanImageInputSpec
    output_spec = _DemeanImageOutputSpec

    def _run_interface(self, runtime):
        from ..utils.images import demean

        self._results["out_file"] = demean(
            self.inputs.in_file,
            self.inputs.in_mask,
            only_mask=self.inputs.only_mask,
            newpath=runtime.cwd,
        )
        return runtime


class _FilledImageLikeInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="image to be demeaned")
    fill_value = traits.Float(1.0, usedefault=True, desc="value to fill")
    dtype = traits.Enum(
        "float32", "uint8", usedefault=True, desc="force output data type"
    )


class _FilledImageLikeOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="demeaned image")


class FilledImageLike(SimpleInterface):
    input_spec = _FilledImageLikeInputSpec
    output_spec = _FilledImageLikeOutputSpec

    def _run_interface(self, runtime):
        from ..utils.images import nii_ones_like

        self._results["out_file"] = nii_ones_like(
            self.inputs.in_file,
            self.inputs.fill_value,
            self.inputs.dtype,
            newpath=runtime.cwd,
        )
        return runtime


class _GenerateSamplingReferenceInputSpec(BaseInterfaceInputSpec):
    fixed_image = File(
        exists=True, mandatory=True, desc="the reference file, defines the FoV"
    )
    moving_image = File(exists=True, mandatory=True, desc="the pixel size reference")
    xform_code = traits.Enum(None, 2, 4, usedefault=True, desc="force xform code")
    fov_mask = traits.Either(
        None,
        File(exists=True),
        usedefault=True,
        desc="mask to clip field of view (in fixed_image space)",
    )
    keep_native = traits.Bool(
        True,
        usedefault=True,
        desc="calculate a grid with native resolution covering "
        "the volume extent given by fixed_image, fast forward "
        "fixed_image otherwise.",
    )


class _GenerateSamplingReferenceOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="one file with all inputs flattened")


class GenerateSamplingReference(SimpleInterface):
    """
    Generates a reference grid for resampling one image keeping original resolution,
    but moving data to a different space (e.g. MNI).

    If the `fov_mask` optional input is provided, then the abbr:`FoV (field-of-view)`
    is cropped to a bounding box containing the brain mask plus an offest of two
    voxels along all dimensions. The `fov_mask` should be to the brain mask calculated
    from the T1w, and should not contain the brain stem. The mask is resampled into
    target space, and then the bounding box is calculated. Finally, the FoV is adjusted
    to that bounding box.


    """

    input_spec = _GenerateSamplingReferenceInputSpec
    output_spec = _GenerateSamplingReferenceOutputSpec

    def _run_interface(self, runtime):
        if not self.inputs.keep_native:
            self._results["out_file"] = self.inputs.fixed_image
            return runtime

        from .. import __version__

        self._results["out_file"] = _gen_reference(
            self.inputs.fixed_image,
            self.inputs.moving_image,
            fov_mask=self.inputs.fov_mask,
            force_xform_code=self.inputs.xform_code,
            message="%s (niworkflows v%s)" % (self.__class__.__name__, __version__),
            newpath=runtime.cwd,
        )
        return runtime


class _IntensityClipInputSpec(BaseInterfaceInputSpec):
    in_file = File(
        exists=True, mandatory=True, desc="3D file which intensity will be clipped"
    )
    p_min = traits.Float(35.0, usedefault=True, desc="percentile for the lower bound")
    p_max = traits.Float(99.98, usedefault=True, desc="percentile for the upper bound")
    nonnegative = traits.Bool(
        True, usedefault=True, desc="whether input intensities must be positive"
    )
    dtype = traits.Enum(
        "int16", "float32", "uint8", usedefault=True, desc="output datatype"
    )
    invert = traits.Bool(False, usedefault=True, desc="finalize by inverting contrast")


class _IntensityClipOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="file after clipping")


class IntensityClip(SimpleInterface):
    """Clip the intensity range as prescribed by the percentiles."""

    input_spec = _IntensityClipInputSpec
    output_spec = _IntensityClipOutputSpec

    def _run_interface(self, runtime):
        self._results["out_file"] = _advanced_clip(
            self.inputs.in_file,
            p_min=self.inputs.p_min,
            p_max=self.inputs.p_max,
            nonnegative=self.inputs.nonnegative,
            dtype=self.inputs.dtype,
            invert=self.inputs.invert,
            newpath=runtime.cwd,
        )
        return runtime


def _gen_reference(
    fixed_image,
    moving_image,
    fov_mask=None,
    out_file=None,
    message=None,
    force_xform_code=None,
    newpath=None,
):
    """Generate a sampling reference, and makes sure xform matrices/codes are correct."""
    import nilearn.image as nli

    if out_file is None:
        out_file = fname_presuffix(
            fixed_image, suffix="_reference", newpath=newpath
        )

    # Moving images may not be RAS/LPS (more generally, transverse-longitudinal-axial)
    reoriented_moving_img = nb.as_closest_canonical(nb.load(moving_image))
    new_zooms = reoriented_moving_img.header.get_zooms()[:3]

    # Avoid small differences in reported resolution to cause changes to
    # FOV. See https://github.com/nipreps/fmriprep/issues/512
    # A positive diagonal affine is RAS, hence the need to reorient above.
    new_affine = np.diag(np.round(new_zooms, 3))

    resampled = nli.resample_img(
        fixed_image, target_affine=new_affine, interpolation="nearest"
    )

    if fov_mask is not None:
        # If we have a mask, resample again dropping (empty) samples
        # out of the FoV.
        fixednii = nb.load(fixed_image)
        masknii = nb.load(fov_mask)

        if np.all(masknii.shape[:3] != fixednii.shape[:3]):
            raise RuntimeError("Fixed image and mask do not have the same dimensions.")

        if not np.allclose(masknii.affine, fixednii.affine, atol=1e-5):
            raise RuntimeError("Fixed image and mask have different affines")

        # Get mask into reference space
        masknii = nli.resample_img(
            masknii, target_affine=new_affine, interpolation="nearest"
        )
        res_shape = np.array(masknii.shape[:3])

        # Calculate a bounding box for the input mask
        # with an offset of 2 voxels per face
        bbox = np.argwhere(np.asanyarray(masknii.dataobj) > 0)
        new_origin = np.clip(bbox.min(0) - 2, a_min=0, a_max=None)
        new_end = np.clip(bbox.max(0) + 2, a_min=0, a_max=res_shape - 1)

        # Find new origin, and set into new affine
        new_affine_4 = resampled.affine.copy()
        new_affine_4[:3, 3] = new_affine_4[:3, :3].dot(new_origin) + new_affine_4[:3, 3]

        # Calculate new shapes
        new_shape = new_end - new_origin + 1
        resampled = nli.resample_img(
            fixed_image,
            target_affine=new_affine_4,
            target_shape=new_shape.tolist(),
            interpolation="nearest",
        )

    xform = resampled.affine  # nibabel will pick the best affine
    _, qform_code = resampled.header.get_qform(coded=True)
    _, sform_code = resampled.header.get_sform(coded=True)

    xform_code = sform_code if sform_code > 0 else qform_code
    if xform_code == 1:
        xform_code = 2

    if force_xform_code is not None:
        xform_code = force_xform_code

    # Keep 0, 2, 3, 4 unchanged
    resampled.header.set_qform(xform, int(xform_code))
    resampled.header.set_sform(xform, int(xform_code))
    resampled.header["descrip"] = "reference image generated by %s." % (
        message or "(unknown software)"
    )
    resampled.to_filename(out_file)
    return out_file


def _advanced_clip(
    in_file, p_min=35, p_max=99.98, nonnegative=True, dtype="int16", invert=False, newpath=None,
):
    """
    Remove outliers at both ends of the intensity distribution and fit into a given dtype.

    This interface tries to emulate ANTs workflows' massaging that truncate images into
    the 0-255 range, and applies percentiles for clipping images.
    For image registration, normalizing the intensity into a compact range (e.g., uint8)
    is generally advised.

    To more robustly determine the clipping thresholds, data are removed of spikes
    with a median filter.
    Once the thresholds are calculated, the denoised data are thrown away and the thresholds
    are applied on the original image.

    """
    from pathlib import Path
    import nibabel as nb
    import numpy as np
    from scipy import ndimage
    from skimage.morphology import ball

    out_file = (Path(newpath or "") / "clipped.nii.gz").absolute()

    # Load data
    img = nb.squeeze_image(nb.load(in_file))
    if len(img.shape) != 3:
        raise RuntimeError(f"<{in_file}> is not a 3D file.")
    data = img.get_fdata(dtype="float32")

    # Calculate stats on denoised version, to preempt outliers from biasing
    denoised = ndimage.median_filter(data, footprint=ball(3))

    a_min = np.percentile(
        denoised[denoised > 0] if nonnegative else denoised,
        p_min
    )
    a_max = np.percentile(
        denoised[denoised > 0] if nonnegative else denoised,
        p_max
    )

    # Clip and cast
    data = np.clip(data, a_min=a_min, a_max=a_max)
    data -= data.min()
    data /= data.max()

    if invert:
        data = 1.0 - data

    if dtype in ("uint8", "int16"):
        data = np.round(255 * data).astype(dtype)

    hdr = img.header.copy()
    hdr.set_data_dtype(dtype)
    img.__class__(data, img.affine, hdr).to_filename(out_file)

    return str(out_file)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Interfaces for handling spaces."""
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterfaceInputSpec,
    SimpleInterface,
)


class _SpaceDataSourceInputSpec(BaseInterfaceInputSpec):
    in_tuple = traits.Tuple(
        (traits.Str, traits.Dict), mandatory=True, desc="a space declaration"
    )


class _SpaceDataSourceOutputSpec(TraitedSpec):
    space = traits.Str(desc="the space identifier, after dropping the cohort modifier.")
    cohort = traits.Str(desc="a cohort specifier")
    resolution = traits.Str(desc="a resolution specifier")
    density = traits.Str(desc="a density specifier")
    uid = traits.Str(desc="a unique identifier combining space specifications")


class SpaceDataSource(SimpleInterface):
    """
    Generate a Nipype interface from a Space specification.

    Example
    -------
    >>> SpaceDataSource(
    ...     in_tuple=('MNIPediatricAsym:cohort-2', {'res': 2, 'den': '91k'})).run().outputs
    <BLANKLINE>
    cohort = 2
    density = 91k
    resolution = 2
    space = MNIPediatricAsym
    uid = MNIPediatricAsym_cohort-2_res-2
    <BLANKLINE>

    >>> SpaceDataSource(
    ...     in_tuple=('MNIPediatricAsym:cohort-2', {'res': 'native', 'den': '91k'})).run().outputs
    <BLANKLINE>
    cohort = 2
    density = 91k
    resolution = native
    space = MNIPediatricAsym
    uid = MNIPediatricAsym_cohort-2_res-native
    <BLANKLINE>

    """

    input_spec = _SpaceDataSourceInputSpec
    output_spec = _SpaceDataSourceOutputSpec

    def _run_interface(self, runtime):
        from ..utils.spaces import format_reference, reference2dict

        self._results = reference2dict(self.inputs.in_tuple)
        self._results["uid"] = format_reference(self.inputs.in_tuple)
        return runtime
